#include "isegclientapi.h"
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>

#define BUF 1024

static void getLine(const char *msg, char *buf, int bufSize)
{
	printf("%s: ", msg);

	fgets(buf, bufSize, stdin);
	char *lf = strchr(buf, '\n');
	if (*lf)
		*lf = 0;
}

static void getItem(const char *name)
{
	char buf[255];
	char *p_pos;
	getLine("Enter item to get", buf, sizeof(buf));

	if (strchr(buf, '*')) {
		uint16_t pos;
		uint8_t separatorNumber = 0;
		uint8_t wild_card = 0;
		for (pos = 0; pos < strlen(buf); pos++) {
			if (buf[pos] == '.')
				separatorNumber++;
			if (buf[pos] == '*') {
				switch (separatorNumber) {
				case 1:
					wild_card++;
					break;

				case 2:
					if (wild_card)
						wild_card++;

					wild_card++;
					break;
				}
			}
		}

		if (separatorNumber == 3) {
			char buf1[255], buf2[255];

			char moduleList[VALUE_SIZE];
			char *p_mod;
			if (wild_card == 3) {
				strncpy(buf1, "0.ModuleList", FULLY_QUALIFIED_OBJECT_SIZE);
				IsegItem item = iseg_getItem(name, buf1);
				strncpy(moduleList, item.value, VALUE_SIZE);
				p_mod = moduleList;
			}

			do {
				strncpy(buf1,buf, FULLY_QUALIFIED_OBJECT_SIZE);

				if (wild_card == 3) {
					p_pos = strchr(buf1, '*');
					*p_pos = *p_mod;
					p_mod += 2;
				}

				strncpy(buf2, buf1, FULLY_QUALIFIED_OBJECT_SIZE);
				p_pos = strrchr(buf2, '*');
				strncpy(p_pos, "ChannelNumber", FULLY_QUALIFIED_OBJECT_SIZE);
				IsegItem item = iseg_getItem(name, buf2);

				uint16_t chhannelNumber = atoi(item.value);
				uint16_t channel;
				for (channel = 0; channel < chhannelNumber; channel++) {
					IsegItem item;
					strncpy(buf2,buf1, FULLY_QUALIFIED_OBJECT_SIZE);
					if (channel < 10)
						*strchr(buf2, '*') = channel + '0';
					else {
						p_pos = strchr(buf2, '*');
						*p_pos = (channel/10 + '0');
						*(p_pos + 1) = channel%10 + '0';
						strncpy(p_pos + 2, strrchr(buf1, '*') + 1, FULLY_QUALIFIED_OBJECT_SIZE);
					}
					item = iseg_getItem(name, buf2);
					printf("'%s': '%s' (Q: %s TR: %s TC: %s)\n", item.object, item.value, item.quality,
						   item.timeStampLastRefreshed, item.timeStampLastChanged);
				}
			} while (wild_card == 3 && *p_mod != 0);
		} else if (separatorNumber == 2 && (p_pos = strchr(buf, '['))) {
			uint16_t index;
			for (index = 0; index < VCC_ARRAY_SIZE; index++) {
				*(p_pos + 1) = (index + '0');
				IsegItem item = iseg_getItem(name, buf);
				printf("'%s': '%s' (Q: %s TR: %s TC: %s)\n", item.object, item.value, item.quality,
					   item.timeStampLastRefreshed, item.timeStampLastChanged);
			}
		}
	} else {
		IsegItem item = iseg_getItem(name, buf);
		printf("'%s': '%s' (Q: %s TR: %s TC: %s)\n", item.object, item.value, item.quality,
			   item.timeStampLastRefreshed, item.timeStampLastChanged);
	}
}

static void setItem(const char *name)
{
	IsegItem item = EmptyIsegItem;
	getLine("Enter item to set", item.object, sizeof(item.object));
	getLine("Enter new value", item.value, sizeof(item.value));

	iseg_setItem(name, item.object, item.value);
}

static void getItemProperty(const char *name)
{
	char buf[255];
	getLine("Enter item to get property", buf, sizeof(buf));

	IsegItemProperty property = iseg_getItemProperty(name, buf);
	printf("Object: '%s' (Q: %s)\n", property.object, property.quality);
	printf("DataType: '%s'\n",property.type);
	printf("Unit: '%s'\n", property.unit);
	printf("Access rights: '%s'\n", property.access);
}

static void getRelease(void)
{
	printf("Version: %s\n", iseg_getVersionString());
}

static void setVoltage(const char *name)
{
	uint8_t slot, channel;
	static float value = 100.f;
	static u_int8_t setOn = 0;

	setOn ^= 1;
	for (slot = 0; slot < 10; slot++) {
		IsegItem item = EmptyIsegItem;
		sprintf(item.object, "0.%d.VoltageRampSpeed", slot);
		sprintf(item.value, "%f", 20.f);
		if (iseg_setItem(name, item.object, item.value) != ISEG_OK) {
			printf("Fehler: %s %s %s", name, item.object, item.value);
		}
		for (channel = 0; channel < 16; channel++) {
			IsegItem item = EmptyIsegItem;
			sprintf(item.object, "0.%d.%d.VoltageSet", slot, channel);
			sprintf(item.value, "%f", value * channel);
			if (iseg_setItem(name, item.object, item.value) != ISEG_OK) {
				printf("Fehler: %s %s %s", name, item.object, item.value);
			}
			value += 1.f;

			sprintf(item.object, "0.%d.%d.Control:3", slot, channel);
			sprintf(item.value, "%d", setOn);
			if (iseg_setItem(name, item.object, item.value) != ISEG_OK) {
				printf("Fehler: %s %s %s", name, item.object, item.value);
			}
		}
		//usleep(200000);
	}
}

int main(void)
{
	const char name[] = "interface1";
	const char interface[] = "can0";

	printf("isegHAL Client Example started\n");
	if (iseg_connect(name, interface, NULL) != ISEG_OK) {
		fprintf(stderr, "Cannot connect to isegHalServer...(%s)\n", strerror(errno));
		return EXIT_FAILURE;
	}
	int quit = 0;
	do {
		printf("Press 'q' [ENTER] to quit, 'g' for get and 's' for set command,\n");
		printf("press 'p' for get item properties.");
		char buf[255];
		fgets(buf, sizeof(buf), stdin);
		switch (toupper(buf[0])) {
		case 'Q':
			quit = 1;
			break;
		case 'G':
			getItem(name);
			break;
		case 'S':
			setItem(name);
			break;
		case 'P':
			getItemProperty(name);
			break;
		case 'V':
			setVoltage(name);
			break;
		case 'R':
			getRelease();
			break;
		}
	} while (!quit);


	iseg_disconnect(name);
	printf("main stopped\n");
	return EXIT_SUCCESS;
}
