#include "isegapi.h"
#include "iseghalcan.h"

#include <QHash>
#include <QMutexLocker>

#include <string.h>

typedef QHash<QString, IsegHal *> IsegHalList;
static IsegHalList m_halList;
static QMutex m_mutex;

static bool isKeyValid(const char *name)
{
	if (!m_halList.contains(QLatin1String(name)))
		return false;

	return true;
}

IsegResult iseg_connect(const char *name, const char *interface, void *reserved)
{
	Q_UNUSED(reserved);

	if (isKeyValid(name))
		return ISEG_WRONG_SESSION_NAME;

	bool firstConnect = true;
	QHashIterator<QString, IsegHal *> i(m_halList);
	QMutexLocker locker(&m_mutex);
	while (i.hasNext()) {
		i.next();
		if (i.value()->getInterface() == QLatin1String(interface)) {
			firstConnect = false;
		}
	}

	IsegHal *hal = new IsegHalCan();

	if (QLatin1String(interface) == "test") {
		m_halList.insert(QLatin1String(name), hal);
		return ISEG_OK;
	}

	if (hal->connect(interface, firstConnect)) {
		delete hal;
		return ISEG_ERROR;
	}

	m_halList.insert(QLatin1String(name), hal);

	return ISEG_OK;
}

IsegResult iseg_disconnect(const char *name)
{
	if (!isKeyValid(name))
		return ISEG_WRONG_SESSION_NAME;

	IsegHal *hal = m_halList.value(name);
	bool disconnect = true;
	QString interfaceForDisconnect = hal->getInterface();
	QHashIterator<QString, IsegHal *> i(m_halList);
	QMutexLocker locker(&m_mutex);
	while (i.hasNext()) {
		i.next();
		if (i.key() == QLatin1String(name))
			continue;

		if (i.value()->getInterface() == interfaceForDisconnect) {
			disconnect = false;
		}
	}

	IsegResult result = ISEG_OK;
	if (disconnect) {
		result = hal->disconnect();
	}

	delete hal;
	m_halList.remove(QLatin1String(name));

	return result;
}

unsigned int iseg_getVersion(void)
{
	const unsigned int version[] = { VER_FILEVERSION };

	return (version[0] << 24) + (version[1] << 16) + (version[2] << 8) + version[3];
}

const char *iseg_getVersionString(void)
{
	return VER_FILEVERSION_STR;
}

IsegResult iseg_setItem(const char *name, const char *object, const char *value)
{
	if (!isKeyValid(name))
		return ISEG_WRONG_SESSION_NAME;

	IsegItem item = EmptyIsegItem;
	strncpy(item.object, object, OBJECT_NAME_SIZE);
	strncpy(item.value, value, VALUE_SIZE);
	IsegHal *hal = m_halList.value(name);
	return hal->setValue(&item);
}

IsegItem iseg_getItem(const char *name, const char *object)
{
	IsegItem item = EmptyIsegItem;

	strncpy(item.quality, ISEG_ITEM_QUALITY_ERROR, QUALITY_SIZE);
	if (!isKeyValid(name))
		return item;

	IsegHal *hal = m_halList.value(name);
	return hal->getValue(object);
}


IsegItemProperty iseg_getItemProperty(const char *name, const char *object)
{
	IsegItemProperty item = EmptyIsegItemProperty;

	if (!isKeyValid(name))
		return item;

	IsegHal *hal = m_halList.value(name);
	return hal->getProperty(object);
}
