#ifndef ISEG_COMMON_H
#define ISEG_COMMON_H

#define INTERFACE_NAME_SIZE 128
#define INTERFACE_TYPE_SIZE 128

#define FULLY_QUALIFIED_OBJECT_SIZE 128
#define OBJECT_NAME_SIZE 32
#define VALUE_SIZE 200
#define DATA_TYPE_SIZE 5
#define UNIT_SIZE 4
#define ACCESS_SIZE 3
#define QUALITY_SIZE 4
#define TIME_SIZE 25

#define WRITE_READ_SIZE 20
#define MAX_READ_MESSAGES 10
#define NO_BIT_ARRAY_INDEX 32
#define NO_FLOAT_ARRAY_INDEX NO_BIT_ARRAY_INDEX + 1

#define ISEG_ITEM_QUALITY_INVALID            "000"
#define ISEG_ITEM_QUALITY_INITIALIZE         "001"
#define ISEG_ITEM_QUALITY_OK                 "002"
#define ISEG_ITEM_QUALITY_COMMUNICATION_BAD  "003"
#define ISEG_ITEM_QUALITY_ERROR              "004"

#define UDS_FILE "/tmp/iseghal.uds"

#ifndef UNREFERENCED_PARAMETER
#	define UNREFERENCED_PARAMETER(P) (void)(P)
#endif

enum {
	DataCacheMaxLines = 16,
	DataCacheMaxCrates = 16,
	DataCacheMaxModules = 64,
	DataCacheMaxChannels = 48
};

typedef enum KSupplyVoltages {
	VCC_POSITIVE_EXTERN = 0,	//+24V
	VCC_NEGATIVE_EXTERN = 1,	//-24V
	VCC_LOGIC_EXTERN = 2,		//5..6V
	VCC_POSITIVE_OPV = 3,		//+12..15V
	VCC_NEGATIVE_OPV = 4,		//-12..-15V
	VCC_LOGIC_5V_INTERN = 5,	//+5V
	VCC_LOGIC_3V_INTERN = 6,	//+3.3V
	VCC_EXTERN_AC_POWER = 7,	//externe AC Power
	VCC_BATTERY = 8,			//UPS
	VCC_ARRAY_SIZE
} KSupplyVoltages;

typedef enum KInternalTemperatures {
	INTERNAL_TEMPERATURE_ARRAY_SIZE = 9
} KInternalTemperatures;

typedef enum IsegLogLevel {
	LogErrors = 1,				//log CAN bus error messages, loss of communication (Stop)
	LogInformation = 2,			//log state of initialising, switch on/off live insertion mode
	LogApiRead = 4,				//log API read accesses (APIRX)
	LogApiWrite = 8,			//log API write accesses (APITX)
	LogCanReceive = 32,			//log CAN receice messages (CANRX)
	LogCanReadRequest = 64,		//log CAN read requests messages (CANRQ)
	LogCanWriteRequest = 128	//log CAN write request messages (CANTX)
} IsegLogLevel;

typedef enum IsegResult {
	ISEG_OK = 0,
	ISEG_ERROR = 1,
	ISEG_WRONG_SESSION_NAME,
	ISEG_WRONG_USER,
	ISEG_WRONG_PASSWORD,
	ISEG_NOT_AUTHORIZED,
	ISEG_NO_SSL_SUPPORT
} IsegResult;

typedef struct IsegItem {
	char object[FULLY_QUALIFIED_OBJECT_SIZE];	//line.module.channel.item  0.0.1.VoltageSet
	char value[VALUE_SIZE];
	char quality[QUALITY_SIZE];
	char timeStampLastRefreshed[TIME_SIZE];
	char timeStampLastChanged[TIME_SIZE];
} IsegItem;
#define EmptyIsegItem { "" ,"" , "", "", ""}

typedef struct IsegItemProperty {
	char object[FULLY_QUALIFIED_OBJECT_SIZE];	//line.module.channel.item  0.0.1.VoltageSet
	char type[DATA_TYPE_SIZE];	// "R4", "UI4", "STR", "BOOL"
	char unit[UNIT_SIZE];		// "V", "A", "C", "%", "%/s", "s"
	char access[ACCESS_SIZE];	// "R", "W", "RW"
	char quality[QUALITY_SIZE];
} IsegItemProperty;
#define EmptyIsegItemProperty { "", "", "", "", "" }

#define TIME_STAMP 1
#define TIME_OUT_ECH4XA_BACK_PLANE 2000
#define TIME_OUT_MODULE_LOG_ON 3000

#define LOG_PATH "/tmp"

#endif // ISEG_COMMON_H

