#include <stdio.h>
#include <string.h>

#include "kstring.h"
#include "ksystem.h"
#include "kline.h"
#include "kcrates_ech4x.h"
#include "kmodules.h"
#include "kisegtime.h"

extern CrateCache m_crateCacheValue[DataCacheMaxLines][DataCacheMaxCrates];
extern CrateCache m_crateEch238CacheValue[DataCacheMaxLines][DataCacheMaxCrates];
extern ModuleCache m_moduleCacheValue[DataCacheMaxLines][DataCacheMaxModules];

CommunicationSystem m_commSystemState = state_init;
static bool m_masterCrate = true;

bool isGroupDataId(uint16_t dataid)
{
	if ((dataid & (EDCP_DATAID_CHANNEL_CODE + EDCP_DATAID_GROUP_CODE)) ==
			EDCP_DATAID_GROUP_CODE) {
		return true;
	}

	return false;
}

void kcan_noMasterCrate(void)
{
	m_masterCrate = false;
}

bool kcan_logged(uint16_t line, uint16_t address)
{
	if (address >= DataCacheMaxModules + (DataCacheMaxCrates << 1))
		return false;

	if (address >= DataCacheMaxModules + DataCacheMaxCrates)
		return m_crateEch238CacheValue[line][address - DataCacheMaxModules - DataCacheMaxCrates].logOn;

	if (address >= DataCacheMaxModules)
		return m_crateCacheValue[line][address - DataCacheMaxModules].logOn;

	if (
			m_moduleCacheValue[line][address].logOn[BoardASingleModule] &&
			m_moduleCacheValue[line][address].downCounterLogOn &&
			m_masterCrate
	) {
		m_moduleCacheValue[line][address].downCounterLogOn--;
		return false;
	}

	return m_moduleCacheValue[line][address].logOn[BoardASingleModule];
}

KCanMessage kcan_setLogOn(uint16_t line, uint16_t address, bool boardB)
{
	uint8_t data[2] = { 1, 0 };
	KCanId id = { 0 };
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	if (address >= DataCacheMaxModules) {
		address -= DataCacheMaxModules;
		if (address >= DataCacheMaxCrates)
			address -= DataCacheMaxCrates;
		id.bit.crate = true;
	}
	id.bit.addressOffset = line;
	id.bit.address = address;
	id.bit.passive = 1;
	KCanMessage msg = kcan_createStandardDcpMessage(id, boardB, extendedId, DCP_DATAID_MOD_LOG_ON, 2, data);

	return msg;
}

KCanMessage kcan_setLogOff(uint16_t line, uint16_t address, bool boardB)
{
	uint8_t data[2] = { 0 };
	KCanId id = { 0 };
	KCanMessage msg = EmptyCanMessage;
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	if (address >= DataCacheMaxModules + DataCacheMaxCrates) {
		address -= (DataCacheMaxModules + DataCacheMaxCrates);
		m_crateEch238CacheValue[line][address].logOn = false;
		id.bit.crate = true;
	} else if (address >= DataCacheMaxModules) {
		address -= DataCacheMaxModules;
		m_crateCacheValue[line][address].logOn = false;
		id.bit.crate = true;
	}

	id.bit.addressOffset = line;
	id.bit.address = address;
	id.bit.passive = 1;
	msg = kcan_createStandardDcpMessage(id, boardB, extendedId, DCP_DATAID_MOD_LOG_ON, 2, data);
	return msg;
}

static KCanMessage kcan_createStdOrExtEdcpMessage(KCanId id, bool boardB, bool extendedId, bool edcp,
											 int dataId, int dlc, const uint8_t * const data)
{
	KCanMessage result = EmptyCanMessage;

	if (!id.bit.crate) {
		if (m_crateCacheValue[0][0].backPlaneType & (1 << id.bit.addressOffset)) {
			id.bit.address = (kcan_isDoubleModule(id.bit.addressOffset, id.bit.address)) ?
						kcan_slotNumberToDoubleModuleAddress(id.bit.address, boardB) :
						kcan_slotNumberToSingleModuleAddress(id.bit.address);
		}
	}

	result.id = id;
	result.bit.extendedId = extendedId;
	result.bit.dlc = dlc;
	result.bit.edcp = edcp;
	result.dataId = dataId;
	result.bit.busy = true;

	int minDlc = edcp ? 2 : 1;
	if (data && dlc > minDlc)
		memrcpy(result.data, data, dlc - minDlc);

	return result;
}

KCanMessage kcan_createStandardEdcpMessage(KCanId id, bool boardB, bool extendedId, int dataId, int dlc, const uint8_t * const data)
{
	return kcan_createStdOrExtEdcpMessage(id, boardB, extendedId, true, dataId, dlc, data);
}

KCanMessage kcan_createStandardDcpMessage(KCanId id, bool boardB, bool extendedId, int dataId, int dlc, const uint8_t * const data)
{
	return kcan_createStdOrExtEdcpMessage(id, boardB, extendedId, false, dataId | DCP_DATAID, dlc, data);
}

bool kcan_isChannelDataId(uint16_t dataid)
{
	if ((dataid & (EDCP_DATAID_CHANNEL_CODE + EDCP_DATAID_GROUP_CODE)) ==
			EDCP_DATAID_CHANNEL_CODE + EDCP_DATAID_GROUP_CODE) {
		return true;
	}

	return false;
}

bool kcan_cacheAllAnwersReceived(uint16_t line, uint16_t address)
{
	if (address >= DataCacheMaxModules + DataCacheMaxCrates) {
		return m_crateEch238CacheValue[line][address - DataCacheMaxModules - DataCacheMaxCrates].timeOut.answers < 1;
	}

	if (address >= DataCacheMaxModules) {
		return m_crateCacheValue[line][address - DataCacheMaxModules].timeOut.answers < 1;
	}

	return m_moduleCacheValue[line][address].timeOut.answers < 1;
}

bool kcan_cacheAlive(uint16_t line, uint16_t address)
{
	if (address >= DataCacheMaxModules + DataCacheMaxCrates) {
		return m_crateEch238CacheValue[line][address - DataCacheMaxModules - DataCacheMaxCrates].alive;
	}

	if (address >= DataCacheMaxModules) {
		return m_crateCacheValue[line][address - DataCacheMaxModules].alive;
	}

	return m_moduleCacheValue[line][address].alive;
}

bool kcan_cacheSetValuesUpdated(uint16_t line, uint16_t address)
{
	if (address <= DataCacheMaxModules) {
		return m_moduleCacheValue[line][address].updated;
	}

	return true;
}

KCanMessage kcan_createAntiCollisionMessage(uint16_t line, uint16_t module, bool boardB)
{
	uint8_t data[2] = { 1, 0 };
	KCanId id = { 0 };
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	id.bit.address = module;
	id.bit.passive = 1;
	KCanMessage msg = kcan_createStandardEdcpMessage(id, boardB, extendedId, EDCP_DATAID_MOD_CONFIGURE_COMM, 4, data);
	return msg;
}

uint8_t kcan_moduleAddressToSlotNumber(uint16_t line, uint16_t address, bool *boardB)
{
	*boardB = false;

	switch (address) {
	case (0x200 - 0x200) >> 3: return 0;
	case (0x208 - 0x200) >> 3: *boardB = true;
							   return 0;

	case (0x210 - 0x200) >> 3:
	case (0x220 - 0x200) >> 3: return 1;
	case (0x228 - 0x200) >> 3: *boardB = true;
							   return 1;

	case (0x218 - 0x200) >> 3:
	case (0x230 - 0x200) >> 3: return 2;
	case (0x238 - 0x200) >> 3: *boardB = true;
							   return 2;

	case (0x260 - 0x200) >> 3:
	case (0x240 - 0x200) >> 3: return 3;
	case (0x248 - 0x200) >> 3: *boardB = true;
							   return 3;

	case (0x268 - 0x200) >> 3:
	case (0x250 - 0x200) >> 3: return 4;
	case (0x258 - 0x200) >> 3: *boardB = true;
							   return 4;

	case (0x270 - 0x200) >> 3: return 5;
	case (0x278 - 0x200) >> 3:
		if (!m_moduleCacheValue[line][5].logOn[BoardBSingleModule1]) {
			*boardB = true;
		}
		return 5;

	case (0x280 - 0x200) >> 3: return 6;
	case (0x288 - 0x200) >> 3: *boardB = true;
							   return 6;

	case (0x290 - 0x200) >> 3:
	case (0x2A0 - 0x200) >> 3: return 7;
	case (0x2A8 - 0x200) >> 3: *boardB = true;
							   return 7;

	case (0x298 - 0x200) >> 3:
	case (0x2B0 - 0x200) >> 3: return 8;
	case (0x2B8 - 0x200) >> 3: *boardB = true;
							   return 8;

	case (0x2E0 - 0x200) >> 3:
	case (0x2C0 - 0x200) >> 3: return 9;
	case (0x2C8 - 0x200) >> 3: *boardB = true;
							   return 9;

	default: return 0;
	}
}

uint16_t kcan_slotNumberToDoubleModuleAddress(uint16_t slot, bool boardB)
{
	switch (slot) {
	case 0: return (!boardB) ? 0 : 1;

	case 1: return (!boardB) ? 4 : 5;

	case 2: return (!boardB) ? 6 : 7;

	case 3: return (!boardB) ? 8 : 9;

	case 4: return (!boardB) ? 10 : 11;

	case 5: return (!boardB) ? 14 : 15;

	case 6: return (!boardB) ? 16 : 17;

	case 7: return (!boardB) ? 20 : 21;

	case 8: return (!boardB) ? 22 : 23;

	case 9: return (!boardB) ? 24 : 25;

	default: return (!boardB) ?  0 : 1;
	}
}

uint16_t kcan_slotNumberToSingleModuleAddress(uint16_t slot)
{
	switch (slot) {
	case 0: return 0;

	case 1: return 2;

	case 2: return 3;

	case 3: return 12;

	case 4: return 13;

	case 5: return 15;

	case 6: return 16;

	case 7: return 18;

	case 8: return 19;

	case 9: return 28;

	default: return 0;
	}
}

bool kcan_isDoubleModule(uint16_t line, uint16_t module)
{
	return m_moduleCacheValue[line][module].logOn[BoardASingleModule] &
					m_moduleCacheValue[line][module].logOn[BoardBModule];
}
