#include "kdatacache.h"
#include "kdatacache_p.h"
#include "kdeviceclass.h"
#include "kedcp_can_dataid.h"
#include "kisegtime.h"
#include "kisegtype.h"
#include "kloggerfifo.h"
#include "kstring.h"

#include <ctype.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

extern SystemCache m_systemCacheValue;
extern char m_systemCacheTimeStampLastRefreshed[DataCacheColumnSystemSize][TIME_SIZE];
extern char m_systemCacheTimeStampLastChanged[DataCacheColumnSystemSize][TIME_SIZE];
extern char m_systemCacheQuality[DataCacheColumnSystemSize][QUALITY_SIZE];
extern ElementState m_systemCacheState[DataCacheColumnSystemSize];
extern const ItemProperty m_systemItemProperties[DataCacheColumnSystemSize];

extern LineCache m_lineCacheValue[DataCacheMaxLines];
extern char m_lineCacheTimeStampLastRefreshed[DataCacheMaxLines][DataCacheColumnLineSize][TIME_SIZE];
extern char m_lineCacheTimeStampLastChanged[DataCacheMaxLines][DataCacheColumnLineSize][TIME_SIZE];
extern char m_lineCacheQuality[DataCacheMaxLines][DataCacheColumnLineSize][QUALITY_SIZE];
extern ElementState m_lineCacheState[DataCacheMaxLines][DataCacheColumnLineSize];
extern const ItemProperty m_lineItemProperties[DataCacheColumnLineSize];

extern CrateCache m_crateCacheValue[DataCacheMaxLines][DataCacheMaxCrates];
extern char m_crateCacheTimeStampLastRefreshed[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate4xaSize + 1][TIME_SIZE];
extern char m_crateCacheTimeStampLastChanged[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate4xaSize + 1][TIME_SIZE];
extern char m_crateCacheQuality[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate4xaSize + 1][QUALITY_SIZE];
extern ElementState m_crateCacheState[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate4xaSize + 2];
extern const ItemProperty m_crateItemProperties[DataCacheColumnCrate4xaSize];

extern CrateCache m_crateEch238CacheValue[DataCacheMaxLines][DataCacheMaxCrates];
extern char m_crateEch238CacheTimeStampLastRefreshed[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate238Size][TIME_SIZE];
extern char m_crateEch238CacheTimeStampLastChanged[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate238Size][TIME_SIZE];
extern char m_crateEch238CacheQuality[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate238Size][QUALITY_SIZE];
extern ElementState m_crateEch238CacheState[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate238Size];
extern const ItemPropertyCrate m_crateEch238ItemProperties[DataCacheColumnCrate238Size];

extern ModuleCache m_moduleCacheValue[DataCacheMaxLines][DataCacheMaxModules];
extern char m_moduleCacheTimeStampLastRefreshed[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize][TIME_SIZE];
extern char m_moduleCacheTimeStampLastChanged[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize][TIME_SIZE];
extern char m_moduleCacheQuality[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize][QUALITY_SIZE];
extern ElementState m_moduleCacheState[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize];
extern const ItemProperty m_moduleItemProperties[DataCacheColumnModuleSize];

extern ItemValue m_channelCache[DataCacheMaxLines][DataCacheMaxModules][DataCacheMaxChannels][DataCacheColumnChannelSize];
extern const ItemProperty m_channelItemProperties[DataCacheColumnChannelSize];

extern CommunicationSystem m_commSystemState;

char m_releaseString[6];
bool m_disconnected = false;
uint32_t m_timeOutBackPlaneType = 0;

static uint32_t initCounter = 0;
static uint32_t preConfiguredCounter = 0;
static KCollectData collectData = CollectEhsModuleData;

void kdatacache_clear(void)
{
	m_commSystemState = state_init;
	m_timeOutBackPlaneType = 0;
	m_disconnected = false;

	initCounter = 0;
	preConfiguredCounter = 0;
	collectData = CollectEhsModuleData;

	memset(m_channelCache, 0, sizeof(m_channelCache));

	memset(m_moduleCacheValue, 0, sizeof(m_moduleCacheValue));
	memset(m_moduleCacheTimeStampLastRefreshed, 0, sizeof(m_moduleCacheTimeStampLastRefreshed));
	memset(m_moduleCacheTimeStampLastChanged, 0, sizeof(m_moduleCacheTimeStampLastChanged));
	memset(m_moduleCacheQuality, 0, sizeof(m_moduleCacheQuality));
	memset(m_moduleCacheState, 0, sizeof(m_moduleCacheState));

	memset(m_crateCacheValue, 0, sizeof(m_crateCacheValue));
	memset(m_crateCacheTimeStampLastRefreshed, 0, sizeof(m_crateCacheTimeStampLastRefreshed));
	memset(m_crateCacheTimeStampLastChanged, 0, sizeof(m_crateCacheTimeStampLastChanged));
	memset(m_crateCacheQuality, 0, sizeof(m_crateCacheQuality));
	memset(m_crateCacheState, 0, sizeof(m_crateCacheState));

	memset(m_crateEch238CacheValue, 0, sizeof(m_crateEch238CacheValue));
	memset(m_crateEch238CacheTimeStampLastRefreshed, 0, sizeof(m_crateEch238CacheTimeStampLastRefreshed));
	memset(m_crateEch238CacheTimeStampLastChanged, 0, sizeof(m_crateEch238CacheTimeStampLastChanged));
	memset(m_crateEch238CacheQuality, 0, sizeof(m_crateEch238CacheQuality));
	memset(m_crateEch238CacheState, 0, sizeof(m_crateEch238CacheState));

	memset(m_lineCacheValue, 0, sizeof(m_lineCacheValue));
	memset(m_lineCacheTimeStampLastRefreshed, 0, sizeof(m_lineCacheTimeStampLastRefreshed));
	memset(m_lineCacheTimeStampLastChanged, 0, sizeof(m_lineCacheTimeStampLastChanged));
	memset(m_lineCacheQuality, 0, sizeof(m_lineCacheQuality));
	memset(m_lineCacheState, 0, sizeof(m_lineCacheState));

	memset(&m_systemCacheValue, 0, sizeof(m_systemCacheValue));
	memset(m_systemCacheTimeStampLastRefreshed, 0, sizeof(m_systemCacheTimeStampLastRefreshed));
	memset(m_systemCacheTimeStampLastChanged, 0, sizeof(m_systemCacheTimeStampLastChanged));
	memset(m_systemCacheQuality, 0, sizeof(m_systemCacheQuality));
	memset(m_systemCacheState, 0, sizeof(m_systemCacheState));

	uint16_t line;
	uint16_t crate;
	uint16_t module;
	uint16_t channel;
	char timeStamp[TIME_SIZE];
	createTimeStamp(timeStamp);
	for (line = 0; line < DataCacheMaxLines; line++) {
		for (module = 0; module < DataCacheMaxModules; module++) {
			m_moduleCacheValue[line][module].downCounterLogOn = 5;
			KDataCacheModuleColumn column;
			for (column = 0; column < DataCacheColumnModuleSize; column++) {
				strncpy(m_moduleCacheQuality[line][module][column], ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE);
				strncpy(m_moduleCacheTimeStampLastRefreshed[line][module][column], timeStamp, TIME_SIZE);
				strncpy(m_moduleCacheTimeStampLastChanged[line][module][column], timeStamp, TIME_SIZE);
			}

			for (channel = 0; channel < DataCacheMaxChannels; channel++) {
				KDataCacheChannelColumn column;
				for (column = 0; column < DataCacheColumnChannelSize; column++) {
					strncpy(m_channelCache[line][module][channel][column].quality, ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE);
					strncpy(m_channelCache[line][module][channel][column].timeStampLastRefreshed, timeStamp, TIME_SIZE);
					strncpy(m_channelCache[line][module][channel][column].timeStampLastChanged, timeStamp, TIME_SIZE);
				}
			}
		}

		for (crate = 0; crate < DataCacheMaxCrates; crate++) {
			KDataCacheCrateColumn column;
			for (column = 0; column < DataCacheColumnCrate238Size; column++) {
				if (column < DataCacheColumnCrate4xaSize) {
					strncpy(m_crateCacheQuality[line][crate][column], ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE);
					strncpy(m_crateCacheTimeStampLastRefreshed[line][crate][column], timeStamp, TIME_SIZE);
					strncpy(m_crateCacheTimeStampLastChanged[line][crate][column], timeStamp, TIME_SIZE);
				}

				strncpy(m_crateEch238CacheQuality[line][crate][column], ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE);
				strncpy(m_crateEch238CacheTimeStampLastRefreshed[line][crate][column], timeStamp, TIME_SIZE);
				strncpy(m_crateEch238CacheTimeStampLastChanged[line][crate][column], timeStamp, TIME_SIZE);
			}
		}

		KDataCacheLineColumn column;
		for (column = 1; column < DataCacheColumnLineSize; column++) {
			strncpy(m_lineCacheQuality[line][column], ISEG_ITEM_QUALITY_INITIALIZE, QUALITY_SIZE);
			strncpy(m_lineCacheTimeStampLastRefreshed[line][column], timeStamp, TIME_SIZE);
			strncpy(m_lineCacheTimeStampLastChanged[line][column], timeStamp, TIME_SIZE);
		}
	}
}

uint32_t kdatacache_getChannelValueUI4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column, uint8_t index)
{
	if (index < NO_BIT_ARRAY_INDEX) {
		return (m_channelCache[line][module][channel][column].value >> index) & 1;
	}

	return m_channelCache[line][module][channel][column].value;
}

uint32_t kdatacache_getChannelValueSetUI4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column, uint8_t index)
{
	if (index < NO_BIT_ARRAY_INDEX) {
		return (m_channelCache[line][module][channel][column].valueSet >> index) & 1;
	}

	return m_channelCache[line][module][channel][column].valueSet;
}

float kdatacache_getChannelValueR4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column)
{

	if (column == DataCacheColumnChannelDelayedTripTime) {
		return (float)m_channelCache[line][module][channel][column].value / 1000.f;
	}

	return toFloat(m_channelCache[line][module][channel][column].value);
}

float kdatacache_getChannelValueSetR4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column)
{

	if (column == DataCacheColumnChannelDelayedTripTime) {
		return (float)m_channelCache[line][module][channel][column].valueSet / 1000.f;
	}

	return toFloat(m_channelCache[line][module][channel][column].valueSet);
}


char *kdatacache_getChannelTimeStampLastRefreshed(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column)
{
	return m_channelCache[line][module][channel][column].timeStampLastRefreshed;
}
char *kdatacache_getChannelTimeStampLastChanged(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column)
{
	return m_channelCache[line][module][channel][column].timeStampLastChanged;
}

char *kdatacache_getChannelQuality(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column)
{
	return m_channelCache[line][module][channel][column].quality;
}

IsegResult kdatacache_setChannelValueUI4(uint32_t value, uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column, uint8_t index)
{
	if (column < DataCacheColumnChannelSize) {
		if (strchr(m_channelItemProperties[column].accessRights, 'W')) {
			m_channelCache[line][module][channel][column].valueSet =
					bitArrayIndexedValue(&m_channelCache[line][module][channel][column].value, value, index);
			m_channelCache[line][module][channel][column].state = newValueLowPriority;
			return ISEG_OK;
		}
	}

	return ISEG_ERROR;
}

IsegResult kdatacache_setChannelValueR4(float value, uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column)
{
	if (column == DataCacheColumnChannelDelayedTripTime) {
		uint32_t value1 = value * 1000.f  + 0.5f;
		return kdatacache_setChannelValueUI4(value1, line, module, channel, column, NO_BIT_ARRAY_INDEX);
	}

	return kdatacache_setChannelValueUI4(toUint(value), line, module, channel, column, NO_BIT_ARRAY_INDEX);
}

uint32_t kdatacache_getModuleValueUI4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	if (column == DataCacheColumnModuleLogOn)
		return m_moduleCacheValue[line][module].logOn[BoardASingleModule];

	if (column == DataCacheColumnModuleAlive)
		return m_moduleCacheValue[line][module].alive;

	if (column == DataCacheColumnModuleDeviceClass)
		return m_moduleCacheValue[line][module].deviceClass;

	if (column == DataCacheColumnModuleStatus) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].status >> index) & 1;
		}

		return m_moduleCacheValue[line][module].status;
	}

	if (column == DataCacheColumnModuleControl) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].control >> index) & 1;
		}

		return m_moduleCacheValue[line][module].control;
	}

	if (column == DataCacheColumnModuleEventStatus) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].eventStatus >> index) & 1;
		}

		return m_moduleCacheValue[line][module].eventStatus;
	}

	if (column == DataCacheColumnModuleEventMask) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].eventMask >> index) & 1;
		}

		return m_moduleCacheValue[line][module].eventMask;
	}

	if (column == DataCacheColumnModuleSampleRate)
		return m_moduleCacheValue[line][module].sampleRate;

	if (column == DataCacheColumnModuleDigitalFilter)
		return m_moduleCacheValue[line][module].digitalFilter;

	if (column == DataCacheColumnModuleSerialNumber)
		return m_moduleCacheValue[line][module].serialNumber;

	if (column == DataCacheColumnModuleChannelNumber)
		return kmodule_channelNumber(line, module);

	return 0;
}

uint32_t kdatacache_getModuleValueSetUI4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	if (column == DataCacheColumnModuleControl) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].controlSet >> index) & 1;
		}

		return m_moduleCacheValue[line][module].controlSet;
	}

	if (column == DataCacheColumnModuleEventStatus) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].eventStatusSet >> index) & 1;
		}

		return m_moduleCacheValue[line][module].eventStatusSet;
	}

	if (column == DataCacheColumnModuleEventMask) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (m_moduleCacheValue[line][module].eventMaskSet >> index) & 1;
		}

		return m_moduleCacheValue[line][module].eventMaskSet;
	}

	if (column == DataCacheColumnModuleSampleRate)
		return m_moduleCacheValue[line][module].sampleRateSet;

	if (column == DataCacheColumnModuleDigitalFilter)
		return m_moduleCacheValue[line][module].digitalFilterSet;

	return 0;
}


float kdatacache_getModuleValueR4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	if (column == DataCacheColumnModuleVoltageRampSpeed)
		return m_moduleCacheValue[line][module].voltageRampSpeed;

	if (column == DataCacheColumnModuleCurrentRampSpeed)
		return m_moduleCacheValue[line][module].currentRampSpeed;

	if (column == DataCacheColumnModuleVoltageLimit)
		return m_moduleCacheValue[line][module].voltageLimit;

	if (column == DataCacheColumnModuleCurrentLimit)
		return m_moduleCacheValue[line][module].currentLimit;

	if (column == DataCacheColumnModuleTemperature)
		return m_moduleCacheValue[line][module].temperature;

	if (column == DataCacheColumnModuleTemperatures)
		return m_moduleCacheValue[line][module].temperatures[index];

	if (column == DataCacheColumnModuleSupplies)
		return m_moduleCacheValue[line][module].supplies[index];

	return NAN;
}

float kdatacache_getModuleValueSetR4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	UNREFERENCED_PARAMETER(index);

	if (column == DataCacheColumnModuleVoltageRampSpeed)
		return m_moduleCacheValue[line][module].voltageRampSpeedSet;

	if (column == DataCacheColumnModuleCurrentRampSpeed)
		return m_moduleCacheValue[line][module].currentRampSpeedSet;

	return NAN;
}

const char *kdatacache_getModuleValueString(uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	if (column == DataCacheColumnModuleFirmwareRelease) {
		m_releaseString[0] = m_moduleCacheValue[line][module].firmwareRelease[3] + '0';
		m_releaseString[1] = m_moduleCacheValue[line][module].firmwareRelease[2] + '0';
		m_releaseString[2] = '.';
		m_releaseString[3] = m_moduleCacheValue[line][module].firmwareRelease[1] + '0';
		m_releaseString[4] = m_moduleCacheValue[line][module].firmwareRelease[0] + '0';
		m_releaseString[5] = 0;
		return m_releaseString;
	}

	if (column == DataCacheColumnModuleFirmwareName)
		return m_moduleCacheValue[line][module].firmwareName;

	if (column == DataCacheColumnModuleHighVoltageOk)
		return m_moduleCacheValue[line][module].highVoltageOk;

	if (column == DataCacheColumnModuleArticle)
		return m_moduleCacheValue[line][module].article;

	if (column == DataCacheColumnModuleIdentifierList)
		return m_moduleCacheValue[line][module].identifierList;

	return NULL;
}

const char *kdatacache_getModuleValueSetString(uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	if (column == DataCacheColumnModuleHighVoltageOk)
		return m_moduleCacheValue[line][module].highVoltageOkSet;

	return NULL;
}

const char *kdatacache_getModuleTimeStampLastRefreshed(uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	return m_moduleCacheTimeStampLastRefreshed[line][module][column];
}

const char *kdatacache_getModuleTimeStampLastChanged(uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	return m_moduleCacheTimeStampLastChanged[line][module][column];
}

const char *kdatacache_getModuleQuality(uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	return m_moduleCacheQuality[line][module][column];
}

static IsegResult setModuleValue(uint32_t value, uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	switch (column) {
	case DataCacheColumnModuleLogOn:
		break;

	case DataCacheColumnModuleControl:
		m_moduleCacheValue[line][module].controlSet =
				bitArrayIndexedValue(&m_moduleCacheValue[line][module].control, value, index);
		break;

	case DataCacheColumnModuleEventStatus:
		m_moduleCacheValue[line][module].eventStatusSet =
				bitArrayIndexedValue(&m_moduleCacheValue[line][module].control, value, index);
		break;

	case DataCacheColumnModuleEventMask:
		m_moduleCacheValue[line][module].eventMaskSet =
				bitArrayIndexedValue(&m_moduleCacheValue[line][module].control, value, index);
		break;

	case DataCacheColumnModuleSampleRate:
		m_moduleCacheValue[line][module].sampleRateSet = value;
		break;

	case DataCacheColumnModuleDigitalFilter:
		m_moduleCacheValue[line][module].digitalFilterSet = value;
		break;

	case DataCacheColumnModuleVoltageRampSpeed:
		m_moduleCacheValue[line][module].voltageRampSpeedSet = toFloat(value);
		break;

	case DataCacheColumnModuleCurrentRampSpeed:
		m_moduleCacheValue[line][module].currentRampSpeedSet = toFloat(value);
		break;

	default:
		return ISEG_ERROR;
	}

	m_moduleCacheState[line][module][column] = newValueLowPriority;
	return ISEG_OK;
}

IsegResult kdatacache_setModuleValueString(const char *string, uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	if (column == DataCacheColumnModuleHighVoltageOk) {
		strncpy(m_moduleCacheValue[line][module].highVoltageOkSet, string, EDCP_MAX_STRING_LENGTH);
		m_moduleCacheState[line][module][column] = newValueLowPriority;
		return ISEG_OK;
	}

	return ISEG_ERROR;
}

IsegResult kdatacache_setModuleValueUI4(uint32_t value, uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	return setModuleValue(value, line, module, column, index);
}

IsegResult kdatacache_setModuleValueR4(float value, uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	return kdatacache_setModuleValueUI4(toUint(value), line, module, column, NO_BIT_ARRAY_INDEX);
}

bool kdatacache_isModulAlive(uint16_t line, uint16_t module)
{
	return m_moduleCacheValue[line][module].alive;
}

//=CrateAccess==================================================================================================
CrateCache *kdatacache_crateCacheValue(uint16_t line, uint16_t crate, bool isEch238)
{
	return (isEch238) ? &m_crateEch238CacheValue[line][crate] : &m_crateCacheValue[line][crate];
}

uint32_t kdatacache_getCrateValueUI4(CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index)
{
	if (column == DataCacheColumnCrateLogOn)
		return crateCache->logOn;

	if (column == DataCacheColumnCrateAlive)
		return crateCache->alive;

	if (column == DataCacheColumnCrateDeviceClass)
		return crateCache->deviceClass;

	if (column == DataCacheColumnCrateStatus) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (crateCache->status >> index) & 1;
		}

		return crateCache->status;
	}

	if (column == DataCacheColumnCrateControl) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (crateCache->control >> index) & 1;
		}

		return crateCache->control;
	}

	if (column == DataCacheColumnCrateEventStatus) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (crateCache->eventStatus >> index) & 1;
		}

		return crateCache->eventStatus;
	}

	if (column == DataCacheColumnCrateEventMask) {
		if (index < NO_BIT_ARRAY_INDEX) {
			return (crateCache->eventMask >> index) & 1;
		}

		return crateCache->eventMask;
	}

	if (column == DataCacheColumnCratePowerOn)
		return crateCache->powerOn;

	if (column == DataCacheColumnCrateSerialNumber)
		return crateCache->serialNumber;


	return 0;
}

float kdatacache_getCrateValueR4(CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index)
{
	if (column == DataCacheColumnCrateFanSpeed)
		return crateCache->fanSpeed;

	if (column == DataCacheColumnCrateTemperatures)
		return crateCache->temperatures[index];

	if (column == DataCacheColumnCrateSupplies)
		return crateCache->supplies[index];

	return NAN;
}

char *kdatacache_getCrateValueString(CrateCache *crateCache, KDataCacheCrateColumn column)
{
	if (column == DataCacheColumnCrateFirmwareRelease) {
		m_releaseString[0] = crateCache->firmwareRelease[3] + '0';
		m_releaseString[1] = crateCache->firmwareRelease[2] + '0';
		m_releaseString[2] = '.';
		m_releaseString[3] = crateCache->firmwareRelease[1] + '0';
		m_releaseString[4] = crateCache->firmwareRelease[0] + '0';
		m_releaseString[5] = 0;
		return m_releaseString;
	}

	if (column == DataCacheColumnCrateFirmwareName)
		return crateCache->firmwareName;

	if (column == DataCacheColumnCrateArticle)
		return crateCache->article;

	if (column == DataCacheColumnCrateIdetifierList) {
		if (!crateCache->alive)
			crateCache->identifierList[0] = 0;

		return crateCache->identifierList;
	}

	return NULL;
}

IsegResult setCrateValue(uint32_t value, CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index)
{
	switch (column) {
	case DataCacheColumnCrateControl:
		crateCache->controlSet = bitArrayIndexedValue(&crateCache->control, value, index);
		break;

	case DataCacheColumnCrateEventStatus:
		crateCache->eventStatusSet = bitArrayIndexedValue(&crateCache->eventStatus, value, index);
		break;

	case DataCacheColumnCrateEventMask:
		crateCache->eventMaskSet = bitArrayIndexedValue(&crateCache->eventMask, value, index);
		break;

	case DataCacheColumnCratePowerOn:
		crateCache->powerOnSet = value;
		break;

	default:
		return ISEG_ERROR;
	}

	return ISEG_OK;
}

IsegResult kdatacache_setCrateValueUI4(uint32_t value, CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index)
{
	return setCrateValue(value, crateCache, column, index);
}

void kdataCache_setCrateCacheStateNewValue(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238)
{
	if (isEch238) {
		m_crateEch238CacheState[line][crate][column] = newValueLowPriority;
	} else {
		m_crateCacheState[line][crate][column] = newValueLowPriority;
	}
}

char *kdatacache_getCrateTimeStampLastRefreshed(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238)
{
	return (isEch238) ? m_crateEch238CacheTimeStampLastRefreshed[line][crate][column] : m_crateCacheTimeStampLastRefreshed[line][crate][column];
}

char *kdatacache_getCrateTimeStampLastChanged(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238)
{
	return (isEch238) ? m_crateEch238CacheTimeStampLastChanged[line][crate][column] : m_crateCacheTimeStampLastChanged[line][crate][column];
}

char *kdatacache_getCrateQuality(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238)
{
	return (isEch238) ? m_crateEch238CacheQuality[line][crate][column] : m_crateCacheQuality[line][crate][column];
}

bool kdatacache_isEch238(uint16_t line, uint16_t crate)
{
	return m_crateEch238CacheValue[line][crate].alive;
}

bool kdatacache_isCrateAlive(uint16_t line, uint16_t crate, bool isEch238)
{
	return (isEch238) ? m_crateEch238CacheValue[line][crate].alive : m_crateCacheValue[line][crate].alive;
}

static int stoi(const char *string, uint8_t length, bool *result)
{
	int value = 0;
	uint8_t index;
	for (index = 0; index < length; index++) {
		value *= 16;
		char c = tolower(string[index]);
		if (c >= '0' && c <='9') value += string[index] - '0';
		else if (c >= 'a' && c <= 'f') {
			value += c - 'a' + 10;
		} else {
			*result = false;
			return value;
		}
	}

	*result  = true;
	return value;
}

static char nibbleToChar(uint8_t byte)
{
	if (byte <= 9)
		return byte + '0';

	return byte - 10 + 'A';
}

static void setCanMessageSystemValue(const KCanMessage *rxMsg)
{
	uint8_t index;
	for (index = 1; index < MAX_READ_MESSAGES; index++)
		if (!strncmp(m_systemCacheValue.read[index], "", WRITE_READ_SIZE)) {
			char *posDataIdDestination;
			char *posDataIdSource = strchr(m_systemCacheValue.read[0], '#') + 1;
			uint8_t length = 4, rxi;
			bool ok;
			snprintf(m_systemCacheValue.read[index], VALUE_SIZE, "%X#", rxMsg->id.w);
			posDataIdDestination = strchr(m_systemCacheValue.read[index], '#') + 1;
			uint16_t dataId = stoi(posDataIdSource, 4, &ok);
			if (dataId & DCP_DATAID) {
				length = 2;
			}
			strncat(posDataIdDestination, posDataIdSource, length);
			posDataIdDestination += length;
			*posDataIdDestination = 0;
			length = rxMsg->bit.dlc - (length >> 1);
			for (rxi = 0; rxi < length; rxi++) {
				*posDataIdDestination ++ = nibbleToChar(rxMsg->data[rxi] >> 4);
				*posDataIdDestination ++ = nibbleToChar(rxMsg->data[rxi] & 0xf);
			}
			return;
		}
}

static bool isCanMessageTunneld(const KCanMessage *rxMsg)
{
	char idStr[10];
	uint8_t length;
	KCanId id;
	bool ok;

	if (!m_systemCacheValue.configuration.bit.modeTunnel)
		return false;

	if (!strncmp(m_systemCacheValue.read[0], "", WRITE_READ_SIZE))
		return false;

	length = strchr(m_systemCacheValue.read[0], '#') - m_systemCacheValue.read[0];
	strncpy(idStr, m_systemCacheValue.read[0], length);
	id.w = stoi(idStr, length, &ok);
	id.bit.readRequest = 0;
	id.bit.antiCollision = 1;
	if (id.w == rxMsg->id.w) {
		uint16_t dataId = stoi(&m_systemCacheValue.read[0][4], 4, &ok);
		if (!ok)
			return false;

		if (!rxMsg->bit.edcp) {
			dataId &= 0xff;
			dataId |= 0x80;
		}

		if (dataId == rxMsg->dataId)
			return true;
	}

	return false;
}

static bool isCanUpdateMessageTunneld(const KCanMessage *rxMsg)
{
	char idStr[10];
	uint8_t length;
	KCanId id;
	bool ok;

	if (!m_systemCacheValue.configuration.bit.modeUpdate)
		return false;

	if (!strncmp(m_systemCacheValue.read[0], "", WRITE_READ_SIZE))
		return false;

	length = strchr(m_systemCacheValue.read[0], '#') - m_systemCacheValue.read[0];
	strncpy(idStr, m_systemCacheValue.read[0], length);
	id.w = stoi(idStr, length, &ok);
	id.bit.readRequest = 0;
	if (id.w == rxMsg->id.w) {
		uint16_t dataId = stoi(&m_systemCacheValue.read[0][4], 4, &ok);
		if (!ok)
			return false;

		if (!rxMsg->bit.edcp) {
			dataId &= 0xff;
			dataId |= 0x80;
		}

		if (dataId == rxMsg->dataId)
			return true;
	}

	return false;
}

static void setCanMessageUpdateValue(const KCanMessage *rxMsg)
{
	char *posDataIdDestination;
	bool ok;
	uint8_t length = 4, rxi;
	if (isCanUpdateMessageTunneld(rxMsg)) {
		char *posDataIdSource = strchr(m_systemCacheValue.read[0], '#') + 1;
		snprintf(m_systemCacheValue.read[1], WRITE_READ_SIZE, "%x#", rxMsg->id.w);
		posDataIdDestination = strchr(m_systemCacheValue.read[1], '#') + 1;
		uint16_t dataId = stoi(posDataIdSource, 4, &ok);
		if (dataId & DCP_DATAID) {
			length = 2;
		}
		strncat(posDataIdDestination, posDataIdSource, length);
		posDataIdDestination += length;
	} else {
		uint8_t index;
		for (index = 2; index < MAX_READ_MESSAGES; index++) {
			if (!strncmp(m_systemCacheValue.read[index], "", WRITE_READ_SIZE)) {
				snprintf(m_systemCacheValue.read[index], WRITE_READ_SIZE, "%x#", rxMsg->id.w);
				posDataIdDestination = strchr(m_systemCacheValue.read[index], '#') + 1;
				if (!(rxMsg->dataId & DCP_DATAID)) {
					*posDataIdDestination++ = nibbleToChar((rxMsg->dataId >> 12) & 0xf);
					*posDataIdDestination++ = nibbleToChar((rxMsg->dataId >> 8) & 0xf);
				} else {
					length = 2;
				}
				*posDataIdDestination++ = nibbleToChar((rxMsg->dataId >> 4) & 0xf);
				*posDataIdDestination++ = nibbleToChar(rxMsg->dataId & 0xf);
			} else {
				return;
			}
		}
	}

	*posDataIdDestination = 0;
	length = rxMsg->bit.dlc - (length >> 1);
	for (rxi = 0; rxi < length; rxi++) {
		*posDataIdDestination++ = nibbleToChar(rxMsg->data[rxi] >> 4);
		*posDataIdDestination++ = nibbleToChar(rxMsg->data[rxi] & 0xf);
	}
}

static bool isCanUpdateMessage(const KCanMessage *rxMsg)
{
	if (!m_systemCacheValue.configuration.bit.modeUpdate)
		return false;

	if ( (rxMsg->dataId & 0xff00) == EDCP_DATAID_MOD_FACTORY_FLASH)
		return true;

	return false;
}

void kdatacache_setCanMessageValue(const KCanMessage *rxMsg)
{
	if (isCanMessageTunneld(rxMsg)) {
		setCanMessageSystemValue(rxMsg);
	} else if (isCanUpdateMessage(rxMsg)) {
		setCanMessageUpdateValue(rxMsg);
	} else if (rxMsg->id.bit.crate) {
		if (
				kcrate_isEch4x(rxMsg->id.bit.addressOffset, rxMsg->id.bit.address) ||
				(rxMsg->dataId == DCP_DATAID + DCP_DATAID_MOD_LOG_ON && rxMsg->data[1] == DC_ECH4XA)
		) {
			kcrate_setCanMessageValue(rxMsg);
		} else {
			kcrateEch238_setCanMessageValue(rxMsg);
		}
	} else {
		kmodule_setCanMessageValue(rxMsg);
	}
}

static KLoggerMessage m_messages[LOGGER_FIFO_CAPACITY];
KLoggerFifo m_fifo;
static bool initializeSystemCache(void)
{
	kloggerfifo_init(&m_fifo, m_messages, LOGGER_FIFO_CAPACITY);
	m_systemCacheValue.logLevel = LogInformation + LogErrors;
	kloggerfifo_setLevel(LogInformation + LogErrors);
	strncpy(m_systemCacheValue.logPath, LOG_PATH, VALUE_SIZE);
	return true;
}

static uint8_t kfix_hexToNibble(char nibble)
{
	if ( (nibble >= '0') && (nibble <= '9') )
		return nibble - '0';

	if ( (nibble >= 'A') && (nibble <= 'F') )
		return nibble - 'A' + 10;

	if ( (nibble >= 'a') && (nibble <= 'f') )
		return nibble - 'a' + 10;

	return 0;
}

static uint8_t kfix_hexToByte(const char *hex)
{
	uint8_t b = kfix_hexToNibble(hex[0]) << 4;
	b |= kfix_hexToNibble(hex[1]);

	return b;
}

KCanMessage canMessageSystemValueString(char *string)
{
	char idStr[10];
	KCanMessage msg = EmptyCanMessage;
	uint8_t length = strchr(string, '#') - string;
	uint8_t index;
	bool ok;

	if (
		length < 3 ||
		length > 8 ||
		!strchr(string, '#')
	) {
		return msg;
	}

	strncpy(idStr, string, length);
	msg.id.w = stoi(idStr, length, &ok);
	if (length > 3)
		msg.bit.extendedId = 1;
	if (ok) {
		msg.bit.dlc = (strlen(string) - length) >> 1;
		if (msg.bit.dlc < 1)
			msg.bit.rtr = 1;
		string += length + 1;
		msg.dataId = kfix_hexToByte(string);
		string += 2;
		for (index = 0; index < msg.bit.dlc - 1; index++, string += 2) {
			msg.data[index] = kfix_hexToByte(string);
		}
		msg.bit.busy = 1;
		//m_systemCacheValue.timeOut.answers = 0;
	}

	return msg;
}

static bool writeSystemData(KCanMessage *msg)
{
	uint16_t column;

	column = DataCacheColumnSystemWrite;
	if (m_systemCacheState[column] == newValueLowPriority) {
		*msg = canMessageSystemValueString(m_systemCacheValue.write);
		memset(m_systemCacheValue.read, 0, sizeof(m_systemCacheValue.read));
		m_systemCacheState[column] = actual;
		return true;
	}

	column = DataCacheColumnSystemRead;
	if (m_systemCacheState[column] == newValueLowPriority) {
		*msg = canMessageSystemValueString(m_systemCacheValue.read[0]);
		memset(m_systemCacheValue.read[1], 0, sizeof(m_systemCacheValue.read) - sizeof(m_systemCacheValue.read[0]));
		m_systemCacheState[column] = actual;
		return true;
	}

	return false;
}

static KCanMessage kdatacache_setBroadCastCanReset(uint16_t line, uint8_t nmtDataId)
{
	KCanMessage msg = EmptyCanMessage;

	msg.id.w = 0x004;
	msg.bit.extendedId = (line) ? 1 : 0;
	msg.id.bit.addressOffset = line;
	msg.bit.busy = 1;
	msg.bit.dlc = 1;
	msg.dataId = nmtDataId;

	return msg;
}

#define		NMT_DID_START			0xc4
#define		NMT_DID_STOP			0xc8
#define		NMT_DID_CAN_RST 		0xcc

static KCanMessage connect(uint32_t *initCounter)
{
	KCanMessage msg = EmptyCanMessage;

	if (++ *initCounter <= 15) {
		msg = kdatacache_setBroadCastCanReset(*initCounter - 1, NMT_DID_STOP);
	} else if ( (*initCounter >= 50 && *initCounter <= 109) && ((*initCounter % 4)==0) ) {
		msg = kdatacache_setBroadCastCanReset((*initCounter - 51) / 4, NMT_DID_CAN_RST);
	} else if (*initCounter >= 140 && *initCounter <= 154) {
		msg = kdatacache_setBroadCastCanReset(*initCounter - 140, NMT_DID_START);
	}

	return msg;
}

static bool isStatePreConfigured(KCanMessage *msg, uint32_t *initCounter)
{
	switch (m_commSystemState) {
		case state_init:
			*msg = connect(initCounter);
			if (*initCounter == 1100) {
				m_commSystemState = state_preConfigured;
			}
			break;

		case state_disconnect_0:
			*initCounter = 0;
			m_commSystemState = state_disconnect_1;
			break;

		case state_disconnect_1:
			*msg = connect(initCounter);
			if (*initCounter == 1200)
				m_disconnected = true;
			break;

		default:
			break;
		}

	return m_commSystemState == state_preConfigured;
}

static bool systemHandler(KCanMessage *msg)
{
	static bool initialize = false;

	if (!initialize) {
		initialize = initializeSystemCache();
		return false;
	}

	return writeSystemData(msg);
}

KCanMessage kdatacache_handler(void)
{
	KCanMessage msg = EmptyCanMessage;

	if (m_timeOutBackPlaneType <= TIME_OUT_MODULE_LOG_ON)
		m_timeOutBackPlaneType++;

	if (isStatePreConfigured(&msg, &initCounter)) {

		if (systemHandler(&msg)) {
			return msg;
		}

		if (!m_systemCacheValue.configuration.bit.modeTunnel && !m_systemCacheValue.configuration.bit.modeUpdate) {
			switch (collectData) {
			case CollectEhsModuleData:
				return kmodule_handler(&preConfiguredCounter, &collectData);

			case CollectEhqModuleData:
				collectData = CollectEhsModuleData;
				return msg;

			case CollectCrateData:
				return kcrate_handler(&preConfiguredCounter, &collectData);

			case Collect238CrateData:
				return kcrateEch238_handler(&preConfiguredCounter, &collectData);

			default:
				m_commSystemState = state_init;
				initCounter = 0;
				preConfiguredCounter = 0;
				collectData = CollectEhsModuleData;
				return msg;
			}
		}
	}

	return msg;
}

KDataCacheChannelColumn kdatacache_getChannelColumn(const char *objectString)
{
	uint16_t column = DataCacheColumnChannelStatus;
	for ( ; column < DataCacheColumnChannelSize; column++) {
		if (strcmp(m_channelItemProperties[column].objectString, objectString) == 0)
			return column;
	}

	return column;
}

KDataCacheModuleColumn kdatacache_getModuleColumn(const char *objectString)
{
	uint16_t column = DataCacheColumnModuleLogOn;
	for ( ; column < DataCacheColumnModuleSize; column++) {
		if (strcmp(m_moduleItemProperties[column].objectString, objectString) == 0)
			return column;
	}

	return column;
}

KDataCacheModuleColumn kdatacache_getCrateColumn(const char *objectString)
{
	uint16_t column = DataCacheColumnCrateLogOn;
	for ( ; column < DataCacheColumnCrate4xaSize; column++) {
		if (strcmp(m_crateItemProperties[column].objectString, objectString) == 0)
			return column;
	}

	return column;
}

KDataCacheLineColumn kdatacache_getDeviceColumn(const char *objectString)
{
	uint16_t column = DataCacheColumnLineCanStatus;
	for ( ; column < DataCacheColumnLineSize; column++) {
		if (strcmp(m_lineItemProperties[column].objectString, objectString) == 0)
			return column;
	}

	return column;
}

DataType kdatacache_getChannelDataTypeEnum(uint16_t column)
{
	return m_channelItemProperties[column].type;
}

const char *kdatacache_getChannelDataTypeString(uint16_t column, uint16_t index)
{
	if (
			index < NO_BIT_ARRAY_INDEX &&
			(column == DataCacheColumnChannelControl || column == DataCacheColumnChannelEventStatus || column == DataCacheColumnChannelEventMask)
	) {
		return m_moduleItemProperties[DataCacheColumnModuleAlive].typeString;
	}

	return m_channelItemProperties[column].typeString;
}

const char *kdatacache_getChannelUnitString(uint16_t column)
{
	return m_channelItemProperties[column].unit;
}

const char *kdatacache_getChannelAccessString(uint16_t column)
{
	return m_channelItemProperties[column].accessRights;
}

DataType kdatacache_getModuleDataTypeEnum(uint16_t column)
{
	return m_moduleItemProperties[column].type;
}

const char *kdatacache_getModuleDataTypeString(uint16_t column, uint16_t index)
{
	if (
			index < NO_BIT_ARRAY_INDEX &&
			(column == DataCacheColumnModuleControl || column == DataCacheColumnModuleEventStatus || column == DataCacheColumnModuleEventMask)
	) {
		return m_moduleItemProperties[DataCacheColumnModuleAlive].typeString;
	}

	return m_moduleItemProperties[column].typeString;
}

const char *kdatacache_getModuleUnitString(uint16_t column)
{
	return m_moduleItemProperties[column].unit;
}

const char *kdatacache_getModuleAccessString(uint16_t column)
{
	return m_moduleItemProperties[column].accessRights;
}

DataType kdatacache_getCrateDataTypeEnum(uint16_t column)
{
	return m_crateItemProperties[column].type;
}

const char *kdatacache_getCrateDataTypeString(uint16_t column, uint16_t index)
{
	if (
			index < NO_BIT_ARRAY_INDEX &&
			(column == DataCacheColumnCrateControl || column == DataCacheColumnCrateEventStatus || column == DataCacheColumnCrateEventMask)
	) {
		return m_moduleItemProperties[DataCacheColumnCrateAlive].typeString;
	}

	return m_crateItemProperties[column].typeString;
}

const char *kdatacache_getCrateUnitString(uint16_t column)
{
	return m_crateItemProperties[column].unit;
}

const char *kdatacache_getCrateAccessString(uint16_t column)
{
	return m_crateItemProperties[column].accessRights;
}

uint32_t kdatacache_getDeviceValue(uint16_t line, KDataCacheLineColumn column)
{
	if (column == DataCacheColumnLineModuleNumber) {
		m_lineCacheValue[line].moduleNumber = kline_moduleNumber(line);
		createTimeStamp(m_lineCacheTimeStampLastRefreshed[line][DataCacheColumnLineModuleNumber]);

		return m_lineCacheValue[line].moduleNumber;
	}

	if (column == DataCacheColumnLineCrateNumber)
		return m_lineCacheValue[line].crateNumber;

	return 0;
}

char *kdatacache_getLineValueString(uint16_t line, KDataCacheLineColumn column, uint8_t index)
{
	UNREFERENCED_PARAMETER(index);

	if (column == DataCacheColumnLineModuleList)
		return m_lineCacheValue[line].moduleList;

	if (column == DataCacheColumnLineCrateList)
		return m_lineCacheValue[line].crateList;

	return NULL;
}

const char *kdatacache_getLinePropertyString(KDataCacheLineColumn column)
{
	if (column < DataCacheColumnLineSize) {
		return m_lineItemProperties[column].objectString;
	}

	return "";
}

const char *kdatacache_getLineTimeStampLastRefreshed(uint16_t line, KDataCacheLineColumn column)
{
	return m_lineCacheTimeStampLastRefreshed[line][column];
}

const char *kdatacache_getLineTimeStampLastChanged(uint16_t line, KDataCacheLineColumn column)
{
	return m_lineCacheTimeStampLastChanged[line][column];
}

bool kdatacache_createLineTimeStamp(uint16_t line, KDataCacheLineColumn column)
{
	if (column != DataCacheColumnLineCanStatus && column != DataCacheColumnLineCanBitRate)
		return false;

	createTimeStamp(m_lineCacheTimeStampLastRefreshed[line][column]);
	return true;
}

const char *kdatacache_getLineQuality(uint16_t line, KDataCacheLineColumn column)
{
	return m_lineCacheQuality[line][column];
}

DataType kdatacache_getLineDataTypeEnum(KDataCacheLineColumn column)
{
	return m_lineItemProperties[column].type;
}

const char *kdatacache_getLineDataTypeString(KDataCacheLineColumn column)
{
	return m_lineItemProperties[column].typeString;
}

const char *kdatacache_getLineUnitString(KDataCacheLineColumn column)
{
	return m_lineItemProperties[column].unit;
}

const char *kdatacache_getLineAccessString(KDataCacheLineColumn column)
{
	return m_lineItemProperties[column].accessRights;
}

uint32_t kdatacache_getSystemValue(KDataCacheSystemColumn column)
{
	if (column == DataCacheColumnSystemCrateNumber) {
		kdatacache_getSystemValueString(DataCacheColumnSystemCrateList, 0);
		return m_systemCacheValue.crateNumber;
	}

	if (column == DataCacheColumnSystemModuleNumber) {
		m_systemCacheValue.moduleNumber = ksystem_moduleNumber();
		createTimeStamp(m_systemCacheTimeStampLastRefreshed[DataCacheColumnSystemModuleNumber]);
		return m_systemCacheValue.moduleNumber;
	}

	if (column == DataCacheColumnSystemCycleCounter)
		return m_systemCacheValue.cycleCounter;

	if (column == DataCacheColumnSystemLogLevel)
		return m_systemCacheValue.logLevel;

	return 0;
}

IsegResult kdatacache_setSystemValue(uint32_t value, KDataCacheSystemColumn column)
{
	if (column == DataCacheColumnSystemConfiguration) {
		m_systemCacheValue.configuration.word = value;
		return ISEG_OK;
	}

	if (column == DataCacheColumnSystemLogLevel) {
		m_systemCacheValue.logLevel = value;
		kloggerfifo_setLevel(value);
		return ISEG_OK;
	}

	return ISEG_ERROR;
}

char *kdatacache_getSystemValueString(KDataCacheSystemColumn column, uint8_t index)
{
	if (column == DataCacheColumnSystemCrateList) {
		uint16_t line;
		uint32_t crateNumber = 0;
		char *crateList = m_systemCacheValue.crateList;
		int byteWritten = 0;
		*crateList = 0;
		for (line = 0; line < DataCacheMaxLines; line++) {
			uint16_t crate;
			crateNumber += m_lineCacheValue[line].crateNumber;

			if (m_lineCacheValue[line].crateList == 0)
				continue;

			for (crate = 0; crate < DataCacheMaxCrates; crate++) {
				if (m_crateCacheValue[line][crate].alive) {
					unsigned long crateNum = CRATE_ADDRESS_OFFSET + crate;
					byteWritten += snprintf(&crateList[byteWritten],
											VALUE_SIZE - byteWritten,
											"%hu.%lu,", line, crateNum);
				}

				if (m_crateEch238CacheValue[line][crate].alive) {
					unsigned long crateNum = CRATE_ECH238_ADDRESS_OFFSET + crate;
					byteWritten += snprintf(&crateList[byteWritten],
											VALUE_SIZE - byteWritten,
											"%hu.%lu,", line, crateNum);
				}
			}
		}
		if (byteWritten)
			crateList[byteWritten - 1] = 0;
		createTimeStamp(m_systemCacheTimeStampLastRefreshed[DataCacheColumnSystemCrateList]);

		m_systemCacheValue.crateNumber = crateNumber;
		createTimeStamp(m_systemCacheTimeStampLastRefreshed[DataCacheColumnSystemCrateNumber]);

		return m_systemCacheValue.crateList;
	}

	if (column == DataCacheColumnSystemRead) {
		return m_systemCacheValue.read[index];
	}

	if (column == DataCacheColumnSystemLogPath) {
		return m_systemCacheValue.logPath;
	}

	return NULL;
}

IsegResult kdatacache_setSystemValueString(const char *string, KDataCacheSystemColumn column)
{
	if (column == DataCacheColumnSystemWrite) {
		strncpy(m_systemCacheValue.write, string, WRITE_READ_SIZE);
		m_systemCacheState[column] = newValueLowPriority;
		return ISEG_OK;
	}
	if (column == DataCacheColumnSystemRead) {
		strncpy(m_systemCacheValue.read[0], string, WRITE_READ_SIZE);
		memset(m_systemCacheValue.read[1], 0, (MAX_READ_MESSAGES - 1) * WRITE_READ_SIZE);
		m_systemCacheState[column] = newValueLowPriority;
		return ISEG_OK;
	}

	if (column == DataCacheColumnSystemLogPath) {
		strncpy(m_systemCacheValue.logPath, string, VALUE_SIZE);
		return ISEG_OK;
	}

	return ISEG_ERROR;
}

KDataCacheSystemColumn kdatacache_getSystemColumn(const char *objectString)
{
	uint16_t column = DataCacheColumnSystemCanStatus;
	for ( ; column < DataCacheColumnSystemSize; column++) {
		if (strcmp(m_systemItemProperties[column].objectString, objectString) == 0)
			return column;
	}

	return column;
}

const char *kdatacache_getSystemTimeStampLastRefreshed(KDataCacheLineColumn column)
{
	return m_systemCacheTimeStampLastRefreshed[column];
}

const char *kdatacache_getSystemTimeStampLastChanged(KDataCacheLineColumn column)
{
	return m_systemCacheTimeStampLastChanged[column];
}

bool kdatacache_createSystemTimeStamp(bool lastChanged, KDataCacheSystemColumn column)
{
	createTimeStamp((lastChanged) ?
						m_systemCacheTimeStampLastChanged[column] :
						m_systemCacheTimeStampLastRefreshed[column]);
	return true;
}

const char *kdatacache_getSystemQuality(KDataCacheLineColumn column)
{
	return m_systemCacheQuality[column];
}

DataType kdatacache_getSystemDataTypeEnum(uint16_t column)
{
	return m_systemItemProperties[column].type;
}

const char *kdatacache_getSystemPropertyString(KDataCacheSystemColumn column)
{
	if (column < DataCacheColumnSystemSize) {
		return m_systemItemProperties[column].objectString;
	}

	return "";
}

const char *kdatacache_getSystemDataTypeString(uint16_t column)
{
	return m_systemItemProperties[column].typeString;
}

const char *kdatacache_getSystemUnitString(uint16_t column)
{
	return m_systemItemProperties[column].unit;
}

const char *kdatacache_getSystemAccessString(uint16_t column)
{
	return m_systemItemProperties[column].accessRights;
}

bool kdatacache_isDisconnected(void)
{
	return m_disconnected;
}

void kdatacache_disconnect(void)
{
	if (
			m_commSystemState != state_disconnect_0 &&
			m_commSystemState != state_disconnect_1
	) {
		m_commSystemState = state_disconnect_0;
	}
}

uint32_t kcache_remainingAnswers(uint16_t line, uint16_t device, bool crate)
{
	if (crate)
		return m_crateCacheValue[line][device].timeOut.answers;

	return m_moduleCacheValue[line][device].timeOut.answers;
}
