#ifndef KDATACACHE_H
#define KDATACACHE_H

#include <stdint.h>

#include "kcan.h"
#include "ksystem.h"
#include "kline.h"
#include "kcrates_ech238.h"
#include "kcrates_ech4x.h"
#include "kmodules.h"
#include "kchannel.h"

#ifdef __cplusplus
	extern "C" {
#endif

void kdatacache_clear(void);
KDataCacheChannelColumn kdatacache_getChannelColumn(const char *objectString);
KDataCacheModuleColumn kdatacache_getModuleColumn(const char *objectString);
KDataCacheModuleColumn kdatacache_getCrateColumn(const char *objectString);
KDataCacheLineColumn kdatacache_getDeviceColumn(const char *objectString);

uint32_t kdatacache_getChannelValueUI4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column, uint8_t index);
float kdatacache_getChannelValueR4(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column);
char *kdatacache_getChannelTimeStampLastRefreshed(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column);
char *kdatacache_getChannelTimeStampLastChanged(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column);
char *kdatacache_getChannelQuality(uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column);
IsegResult kdatacache_setChannelValueUI4(uint32_t value, uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column, uint8_t index);
IsegResult kdatacache_setChannelValueR4(float value, uint16_t line, uint16_t module, uint16_t channel, KDataCacheChannelColumn column);

uint32_t kdatacache_getModuleValueUI4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index);
float kdatacache_getModuleValueR4(uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index);
const char *kdatacache_getModuleTimeStampLastRefreshed(uint16_t line, uint16_t module, KDataCacheModuleColumn column);
const char *kdatacache_getModuleTimeStampLastChanged(uint16_t line, uint16_t module, KDataCacheModuleColumn column);
const char *kdatacache_getModuleQuality(uint16_t line, uint16_t module, KDataCacheModuleColumn column);
IsegResult kdatacache_setModuleValueString(const char *string, uint16_t line, uint16_t module, KDataCacheModuleColumn column);
IsegResult kdatacache_setModuleValueUI4(uint32_t value, uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index);
const char *kdatacache_getModuleValueString(uint16_t line, uint16_t module, KDataCacheModuleColumn column);
IsegResult kdatacache_setModuleValueR4(float value, uint16_t line, uint16_t module, KDataCacheModuleColumn column);
bool kdatacache_isModulAlive(uint16_t line, uint16_t module);

CrateCache *kdatacache_crateCacheValue(uint16_t line, uint16_t crate, bool isEch238);
uint32_t kdatacache_getCrateValueUI4(CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index);
IsegResult kdatacache_setCrateValueUI4(uint32_t value, CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index);
void kdataCache_setCrateCacheStateNewValue(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238);
float kdatacache_getCrateValueR4(CrateCache *crateCache, KDataCacheCrateColumn column, uint8_t index);
char *kdatacache_getCrateValueString(CrateCache *crateCache, KDataCacheCrateColumn column);
char *kdatacache_getCrateTimeStampLastRefreshed(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238);
char *kdatacache_getCrateTimeStampLastChanged(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238);
char *kdatacache_getCrateQuality(uint16_t line, uint16_t crate, KDataCacheCrateColumn column, bool isEch238);
bool kdatacache_isEch238(uint16_t line, uint16_t crate);
bool kdatacache_isCrateAlive(uint16_t line, uint16_t crate, bool isEch238);

void kdatacache_setCanMessageValue(const KCanMessage *rxMsg);
KCanMessage kdatacache_handler(void);

DataType kdatacache_getChannelDataTypeEnum(uint16_t column);
const char *kdatacache_getChannelDataTypeString(uint16_t column, uint16_t index);
const char *kdatacache_getChannelUnitString(uint16_t column);
const char *kdatacache_getChannelAccessString(uint16_t column);

DataType kdatacache_getModuleDataTypeEnum(uint16_t column);
const char *kdatacache_getModuleDataTypeString(uint16_t column, uint16_t index);
const char *kdatacache_getModuleUnitString(uint16_t column);
const char *kdatacache_getModuleAccessString(uint16_t column);

DataType kdatacache_getCrateDataTypeEnum(uint16_t column);
const char *kdatacache_getCrateDataTypeString(uint16_t column, uint16_t index);
const char *kdatacache_getCrateUnitString(uint16_t column);
const char *kdatacache_getCrateAccessString(uint16_t column);

uint32_t kdatacache_getDeviceValue(uint16_t line, KDataCacheLineColumn column);
char *kdatacache_getLineValueString(uint16_t line, KDataCacheLineColumn column, uint8_t index);
const char *kdatacache_getLineTimeStampLastRefreshed(uint16_t line, KDataCacheLineColumn column);
const char *kdatacache_getLineTimeStampLastChanged(uint16_t line, KDataCacheLineColumn column);
bool kdatacache_createLineTimeStamp(uint16_t line, KDataCacheLineColumn column);
const char *kdatacache_getLineQuality(uint16_t line, KDataCacheLineColumn column);
DataType kdatacache_getLineDataTypeEnum(KDataCacheLineColumn column);
const char *kdatacache_getLinePropertyString(KDataCacheLineColumn column);
const char *kdatacache_getLineDataTypeString(KDataCacheLineColumn column);
const char *kdatacache_getLineUnitString(KDataCacheLineColumn column);
const char *kdatacache_getLineAccessString(KDataCacheLineColumn column);

uint32_t kdatacache_getSystemValue(KDataCacheSystemColumn column);
IsegResult kdatacache_setSystemValue(uint32_t value, KDataCacheSystemColumn column);
char *kdatacache_getSystemValueString(KDataCacheSystemColumn column, uint8_t index);
IsegResult kdatacache_setSystemValueString(const char *string, KDataCacheSystemColumn column);
KDataCacheSystemColumn kdatacache_getSystemColumn(const char *objectString);
const char *kdatacache_getSystemTimeStampLastRefreshed(KDataCacheLineColumn column);
const char *kdatacache_getSystemTimeStampLastChanged(KDataCacheLineColumn column);
bool kdatacache_createSystemTimeStamp(bool lastChanged, KDataCacheSystemColumn column);
const char *kdatacache_getSystemQuality(KDataCacheLineColumn column);
DataType kdatacache_getSystemDataTypeEnum(uint16_t column);
const char *kdatacache_getSystemPropertyString(KDataCacheSystemColumn column);
const char *kdatacache_getSystemDataTypeString(uint16_t column);
const char *kdatacache_getSystemUnitString(uint16_t column);
const char *kdatacache_getSystemAccessString(uint16_t column);

bool kdatacache_isDisconnected(void);
void kdatacache_disconnect(void);

uint32_t kcache_remainingAnswers(uint16_t line, uint16_t device, bool crate);

#ifdef __cplusplus
	}
#endif

#endif // KDATACACHE_H
