#include "kchannel.h"
#include "kcrates_ech4x.h"
#include "kdeviceclass.h"
#include "kisegtime.h"
#include "kline.h"
#include "kloggerfifo.h"
#include "kmodules.h"
#include "kstring.h"
#include "ksystem.h"

#include <pthread.h>

#include <string.h>
#include <stdlib.h>
#include <stdio.h>

static pthread_mutex_t m_mutexAnswers = PTHREAD_MUTEX_INITIALIZER;
static uint16_t m_line = DataCacheMaxLines;

ModuleCache m_moduleCacheValue[DataCacheMaxLines][DataCacheMaxModules] = { 0 }; //EmptyModuleCache;
char m_moduleCacheTimeStampLastRefreshed[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize][TIME_SIZE] = { { { 0 } } };
char m_moduleCacheTimeStampLastChanged[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize][TIME_SIZE] = { { { 0 } } };
char m_moduleCacheQuality[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize][QUALITY_SIZE] = { { { 0 } } };
ElementState m_moduleCacheState[DataCacheMaxLines][DataCacheMaxModules][DataCacheColumnModuleSize] = { { 0 } };
const ItemProperty m_moduleItemProperties[DataCacheColumnModuleSize] = {
   { "Connected",        DCP_DATAID_MOD_LOG_ON,                      DataCacheColumnModuleLogOn,              DataTypeBool,   "BOOL","",    "RW"},
   { "Alive",            DCP_DATAID_MOD_LOG_ON,                      DataCacheColumnModuleAlive,              DataTypeBool,   "BOOL","",    "R" },
   { "Status",           EDCP_DATAID_MOD_STATUS,                     DataCacheColumnModuleStatus,             DataTypeUI4,    "UI4", "",    "R" },
   { "Control",          EDCP_DATAID_MOD_CONTROL,                    DataCacheColumnModuleControl,            DataTypeUI4,    "UI4", "",    "RW"},
   { "EventStatus",      EDCP_DATAID_MOD_EVENT_STATUS,               DataCacheColumnModuleEventStatus,        DataTypeUI4,    "UI4", "",    "RW"},
   { "EventMask",        EDCP_DATAID_MOD_EVENT_MASK,                 DataCacheColumnModuleEventMask,          DataTypeUI4,    "UI4", "",    "RW"},
   { "SampleRate",       EDCP_DATAID_MOD_ADC_SAMPLES_PER_SECOND,     DataCacheColumnModuleSampleRate,         DataTypeUI4,    "UI4", "SPS", "RW"},
   { "DigitalFilter",    EDCP_DATAID_MOD_DIGITAL_FILTER,             DataCacheColumnModuleDigitalFilter,      DataTypeUI4,    "UI4", "",    "RW"},
   { "VoltageRampSpeed", EDCP_DATAID_MOD_VOLTAGE_RAMP_SPEED,         DataCacheColumnModuleVoltageRampSpeed,   DataTypeR4,     "R4",  "%/s", "RW"},
   { "CurrentRampSpeed", EDCP_DATAID_MOD_CURRENT_RAMP_SPEED,         DataCacheColumnModuleCurrentRampSpeed,   DataTypeR4,     "R4",  "%/s", "RW"},
   { "VoltageLimit",     EDCP_DATAID_MOD_VOLTAGE_LIMIT_REL,          DataCacheColumnModuleVoltageLimit,       DataTypeR4,     "R4",  "%",   "R" },
   { "CurrentLimit",     EDCP_DATAID_MOD_CURRENT_LIMIT_REL,          DataCacheColumnModuleCurrentLimit,       DataTypeR4,     "R4",  "%",   "R" },
   { "SupplyP5",         EDCP_DATAID_MOD_SUPPLY5,                    DataCacheColumnModuleSupplyP5,           DataTypeR4,     "R4",  "V",   "R" },
   { "SupplyP24",        EDCP_DATAID_MOD_SUPPLY24,                   DataCacheColumnModuleSupplyP24,          DataTypeR4,     "R4",  "V",   "R" },
   { "Temperature",      EDCP_DATAID_MOD_TEMPERATURE,                DataCacheColumnModuleTemperature,        DataTypeR4,     "R4",  "C",   "R" },
   { "SerialNumber",     EDCP_DATAID_MOD_SERIAL_NUMBER,              DataCacheColumnModuleSerialNumber,       DataTypeUI4,    "UI4", "",    "R" },
   { "ChannelNumber",    EDCP_DATAID_MOD_CHANNEL_NUMBER,             DataCacheColumnModuleChannelNumber,      DataTypeUI4,    "UI4", "",    "R" },
   { "DeviceClass",      DCP_DATAID_MOD_LOG_ON,                      DataCacheColumnModuleDeviceClass,        DataTypeUI4,    "UI4", "",    "R" },
   { "FirmwareRelease",  EDCP_DATAID_MOD_FIRMWARE_RELEASE,           DataCacheColumnModuleFirmwareRelease,    DataTypeString, "STR", "",    "R" },
   { "FirmwareName",     EDCP_DATAID_MOD_FIRMWARE_NAME,              DataCacheColumnModuleFirmwareName,       DataTypeString, "STR", "",    "R" },
   { "Temperatures",     EDCP_DATAID_GRP_TEMPERATURE_SENSOR,         DataCacheColumnModuleTemperatures,       DataTypeR4,     "R4",  "C",   "R" },
   { "Supplies",         EDCP_DATAID_GRP_SUPPLY_MEASURE,             DataCacheColumnModuleSupplies,           DataTypeR4,     "R4",  "V",   "R" },
   { "HighVoltageOk",    EDCP_DATAID_MOD_CONFIGURE_HIGH_VOLTAGE_OK,  DataCacheColumnModuleHighVoltageOk,      DataTypeString, "STR", "",    "RW"},
   { "Article",          EDCP_DATAID_MOD_ARTICLE_DESCRIPTION,        DataCacheColumnModuleArticle,            DataTypeString, "STR", "",    "R" },
   { "IdentifierList",   DCP_DATAID_MOD_LOG_ON,                      DataCacheColumnModuleIdentifierList,      DataTypeString, "STR", "",    "R" }
};

const ItemProperty m_moduleItemExtendedProperties[ExtendedColumnModuleSize - DataCacheColumnModuleSize] = {
   { "Control2",         EDCP_DATAID_MOD_CONTROL2,                   ExtendedColumnModuleControl2,            DataTypeUI4,    "UI4", "",    "RW"},
   { "Status2",          EDCP_DATAID_MOD_STATUS2,                    ExtendedColumnModuleStatus2,             DataTypeUI4,    "UI4", "",    "R" },
   { "EventStatus2",     EDCP_DATAID_MOD_EVENT_STATUS2,              ExtendedColumnModuleEventStatus2,        DataTypeUI4,    "UI4", "",    "RW"},
   { "LogOnBoard2",      DCP_DATAID_MOD_LOG_ON,                      ExtendedColumnModuleDeviceClass2,        DataTypeBool,   "BOOL","",    "R" },
   { "SetValuesChanged", EDCP_DATAID_MOD_COUNTER_SET_VALUES_CHANGED, ExtendedColumnModuleSetValuesChanged,    DataTypeUI4,    "UI4", "",    "R" }
};

extern uint32_t m_timeOutBackPlaneType;

extern CrateCache m_crateCacheValue[DataCacheMaxLines][DataCacheMaxCrates];
extern ElementState m_crateCacheState[DataCacheMaxLines][DataCacheMaxCrates][DataCacheColumnCrate4xaSize + 2];

extern SystemCache m_systemCacheValue;
extern char m_systemCacheTimeStampLastRefreshed[DataCacheColumnSystemSize][TIME_SIZE];
extern char m_systemCacheTimeStampLastChanged[DataCacheColumnSystemSize][TIME_SIZE];
extern char m_systemCacheQuality[DataCacheColumnSystemSize][QUALITY_SIZE];

extern LineCache m_lineCacheValue[DataCacheMaxLines];
extern char m_lineCacheTimeStampLastRefreshed[DataCacheMaxLines][DataCacheColumnLineSize][TIME_SIZE];
extern char m_lineCacheTimeStampLastChanged[DataCacheMaxLines][DataCacheColumnLineSize][TIME_SIZE];
extern char m_lineCacheQuality[DataCacheMaxLines][DataCacheColumnLineSize][QUALITY_SIZE];

extern ItemValue m_channelCache[DataCacheMaxLines][DataCacheMaxModules][DataCacheMaxChannels][DataCacheColumnChannelSize];

uint32_t kdatacache_getSystemValue(KDataCacheSystemColumn column);
void writeLogEvent(const char *messageType, uint32_t level, uint16_t answers, int count);

inline float toFloat(uint32_t value)
{
	FloatInt fi;
	fi.uintVal = value;
	return fi.floatVal;
}

inline uint32_t toUint(float value)
{
	FloatInt fi;
	fi.floatVal = value;
	return fi.uintVal;
}

static void kmodule_setReceivedValue(uint32_t value, uint16_t line, uint16_t module, KDataCacheModuleColumn column, uint8_t index)
{
	bool valueChanged = false;

	switch (column) {
	case DataCacheColumnModuleLogOn:
		if (m_moduleCacheValue[line][module].logOn[BoardASingleModule]) {
			m_moduleCacheValue[line][module].logOn[BoardASingleModule] = false;
			valueChanged = true;
		}
		m_moduleCacheValue[line][module].timeOut.cacheState = state_logOff;
		break;

	case DataCacheColumnModuleDeviceClass:
	case DataCacheColumnModuleIdentifierList:
		if (m_moduleCacheValue[line][module].deviceClass != value) {
			m_moduleCacheValue[line][module].deviceClass = value;
			strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleLogOn], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
			valueChanged = true;
		}
		m_moduleCacheValue[line][module].logOn[BoardASingleModule] = true;
		m_moduleCacheState[line][module][DataCacheColumnModuleLogOn] = actual;
		break;

	case DataCacheColumnModuleStatus:
		if ( (m_moduleCacheValue[line][module].status & 0xffff) != value) {
			m_moduleCacheValue[line][module].status = (m_moduleCacheValue[line][module].status & 0xffff0000) + value;
			valueChanged = true;
		}
		break;

	case ExtendedColumnModuleStatus2:
		if ( (m_moduleCacheValue[line][module].status >> 16) != value) {
			m_moduleCacheValue[line][module].status = (value << 16) + (m_moduleCacheValue[line][module].status & 0xffff);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleControl:
		if (m_moduleCacheValue[line][module].control != value) {
			m_moduleCacheValue[line][module].control = value;
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleEventStatus:
		if (m_moduleCacheValue[line][module].eventStatus != value) {
			m_moduleCacheValue[line][module].eventStatus = value;
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleEventMask:
		if (m_moduleCacheValue[line][module].eventMask != value) {
			m_moduleCacheValue[line][module].eventMask = value;
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleSampleRate:
		if (m_moduleCacheValue[line][module].sampleRate != value) {
			m_moduleCacheValue[line][module].sampleRate = value;
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleDigitalFilter:
		if (m_moduleCacheValue[line][module].digitalFilter != value) {
			m_moduleCacheValue[line][module].digitalFilter = value;
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleSerialNumber:
		if (m_moduleCacheValue[line][module].serialNumber != value) {
			m_moduleCacheValue[line][module].serialNumber = value;
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleChannelNumber:
		if (m_moduleCacheValue[line][module].channelNumber != value) {
			m_moduleCacheValue[line][module].channelNumber = value;
			if (value <= 16) {	//Kanalmaske zur Abfrage multipler Kanaleigenschaften z.B. VoltageSet 0x6100
				m_moduleCacheValue[line][module].channelMask = (1 << value) - 1;
			}
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleFirmwareRelease:
		if (memcmp(m_moduleCacheValue[line][module].firmwareRelease, &value, 4)) {
			memcpy(m_moduleCacheValue[line][module].firmwareRelease, &value, 4);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleVoltageRampSpeed:
		if (m_moduleCacheValue[line][module].voltageRampSpeed != toFloat(value)) {
			m_moduleCacheValue[line][module].voltageRampSpeed = toFloat(value);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleCurrentRampSpeed:
		if (m_moduleCacheValue[line][module].currentRampSpeed != toFloat(value)) {
			m_moduleCacheValue[line][module].currentRampSpeed = toFloat(value);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleVoltageLimit:
		if (m_moduleCacheValue[line][module].voltageLimit != toFloat(value)) {
			m_moduleCacheValue[line][module].voltageLimit = toFloat(value);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleCurrentLimit:
		if (m_moduleCacheValue[line][module].currentLimit != toFloat(value)) {
			m_moduleCacheValue[line][module].currentLimit = toFloat(value);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleSupplyP5:
		if (m_moduleCacheValue[line][module].supplies[VCC_LOGIC_EXTERN] != toFloat(value)) {
			m_moduleCacheValue[line][module].supplies[VCC_LOGIC_EXTERN] = toFloat(value);
			valueChanged = true;
		}
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleSupplies]);
		strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleSupplies], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		m_moduleCacheState[line][module][DataCacheColumnModuleSupplies] = actual;
		break;

	case DataCacheColumnModuleSupplyP24:
		if (m_moduleCacheValue[line][module].supplies[VCC_POSITIVE_EXTERN] != toFloat(value)) {
			m_moduleCacheValue[line][module].supplies[VCC_POSITIVE_EXTERN] = toFloat(value);
			valueChanged = true;
		}
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleSupplies]);
		strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleSupplies], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		m_moduleCacheState[line][module][DataCacheColumnModuleSupplies] = actual;
		break;

	case DataCacheColumnModuleTemperature:
		if (m_moduleCacheValue[line][module].temperature != toFloat(value)) {
			m_moduleCacheValue[line][module].temperature = toFloat(value);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleTemperatures:
		if (
				index < INTERNAL_TEMPERATURE_ARRAY_SIZE &&
				m_moduleCacheValue[line][module].temperatures[index] != toFloat(value)
		) {
			m_moduleCacheValue[line][module].temperatures[index] = toFloat(value);
			valueChanged = true;
		}
		break;

	case DataCacheColumnModuleSupplies:
		if (
				index < VCC_ARRAY_SIZE &&
				m_moduleCacheValue[line][module].supplies[index] != toFloat(value)
		) {
			m_moduleCacheValue[line][module].supplies[index] = toFloat(value);
			valueChanged = true;
		}
		break;

	case ExtendedColumnModuleDeviceClass2:
		if (
				m_moduleCacheValue[line][module].logOn[BoardASingleModule] &&
				(!m_moduleCacheValue[line][module].logOn[BoardBSingleModule1])
		) {
			m_moduleCacheValue[line][module].logOn[BoardBModule] = true;
			break;
		}

		if (m_timeOutBackPlaneType < TIME_OUT_ECH4XA_BACK_PLANE) {
			return;
		}

		m_moduleCacheValue[line][module].deviceClass = value;
		valueChanged = true;

		if (!m_moduleCacheValue[line][module].logOn[BoardBSingleModule2]) {
			m_moduleCacheValue[line][module].logOn[BoardBSingleModule2] = true;
			m_timeOutBackPlaneType = TIME_OUT_ECH4XA_BACK_PLANE;
			return;
		}
		m_moduleCacheValue[line][module].logOn[BoardASingleModule] = true;
		m_moduleCacheValue[line][module].logOn[BoardBSingleModule1] = true;
		m_moduleCacheState[line][module][DataCacheColumnModuleLogOn] = actual;
		break;

	case ExtendedColumnModuleSetValuesChanged:
		if (m_moduleCacheValue[line][module].counterSetValuesChanged < value) {
			m_moduleCacheState[line][module][DataCacheColumnModuleControl] =
			m_moduleCacheState[line][module][DataCacheColumnModuleEventMask] =
			m_moduleCacheState[line][module][DataCacheColumnModuleSampleRate] =
			m_moduleCacheState[line][module][DataCacheColumnModuleDigitalFilter] =
			m_moduleCacheState[line][module][DataCacheColumnModuleVoltageRampSpeed] =
			m_moduleCacheState[line][module][DataCacheColumnModuleCurrentRampSpeed] =
			m_moduleCacheState[line][module][DataCacheColumnModuleVoltageLimit] =
			m_moduleCacheState[line][module][DataCacheColumnModuleCurrentLimit] =
			m_channelCache[line][module][0][DataCacheColumnChannelVoltageSet].state =
			m_channelCache[line][module][0][DataCacheColumnChannelCurrentSet].state =
			m_channelCache[line][module][0][DataCacheColumnChannelVoltageBounds].state =
			m_channelCache[line][module][0][DataCacheColumnChannelCurrentBounds].state =
			m_channelCache[line][module][0][DataCacheColumnChannelControl].state =
			m_channelCache[line][module][0][DataCacheColumnChannelTemperatureExternal].state =
			m_channelCache[line][module][0][DataCacheColumnChannelVctCoefficient].state =
			m_channelCache[line][module][0][DataCacheColumnChannelEventMask].state = waitForInitialData;
			if (m_moduleCacheValue[line][module].serialNumber > 999999) {
				m_channelCache[line][module][0][DataCacheColumnChannelDelayedTripAction].state =
				m_channelCache[line][module][0][DataCacheColumnChannelDelayedTripTime].state =
				m_channelCache[line][module][0][DataCacheColumnChannelExternalInhibitAction].state = waitForInitialData;
			}

			m_moduleCacheValue[line][module].updated = false;
			m_moduleCacheValue[line][module].timeOut.cacheState = state_uninitialized;
		}

		m_moduleCacheValue[line][module].counterSetValuesChanged = value;
		break;

	default:
		return;
	}

	if (column < DataCacheColumnModuleSize) {
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][column]);
		if (valueChanged) {
			strncpy(m_moduleCacheTimeStampLastChanged[line][module][column],
					m_moduleCacheTimeStampLastRefreshed[line][module][column], TIME_SIZE);
		}
		strncpy(m_moduleCacheQuality[line][module][column], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		m_moduleCacheState[line][module][column] = actual;
	}
}

static void kmodule_setReceivedStringValue(const char *string, uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	if (column == DataCacheColumnModuleFirmwareName) {
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleFirmwareName]);
		if (strncmp(m_moduleCacheValue[line][module].firmwareName, string, EDCP_MAX_STRING_LENGTH)) {
			strncpy(m_moduleCacheValue[line][module].firmwareName, string, EDCP_MAX_STRING_LENGTH);
			strncpy(m_moduleCacheTimeStampLastChanged[line][module][DataCacheColumnModuleFirmwareName],
					m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleFirmwareName], TIME_SIZE);
		}
		strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleFirmwareName], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		m_moduleCacheState[line][module][DataCacheColumnModuleFirmwareName] = actual;
	}

	if (column == DataCacheColumnModuleHighVoltageOk) {
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleHighVoltageOk]);
		if (strncmp(m_moduleCacheValue[line][module].highVoltageOk, string, EDCP_MAX_STRING_LENGTH)) {
			strncpy(m_moduleCacheValue[line][module].highVoltageOk, string, EDCP_MAX_STRING_LENGTH);
			strncpy(m_moduleCacheTimeStampLastChanged[line][module][DataCacheColumnModuleHighVoltageOk],
					m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleHighVoltageOk], TIME_SIZE);
		}
		strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleHighVoltageOk], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		m_moduleCacheState[line][module][DataCacheColumnModuleHighVoltageOk] = actual;
	}
}

static void kmodule_setReceivedStringGroupValue(const char *data, uint8_t length, uint16_t line, uint16_t module, KDataCacheModuleColumn column)
{
	if (
			column == DataCacheColumnModuleArticle &&
			(data[0] * EDCP_SINGLE_GROUP_FRAME_LENGTH + length < EDCP_GROUP_MAX_STRING_LENGTH)
	) {
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleArticle]);
		strncpy(&m_moduleCacheValue[line][module].article[data[0] * EDCP_SINGLE_GROUP_FRAME_LENGTH], &data[1], length);
		strncpy(m_moduleCacheTimeStampLastChanged[line][module][DataCacheColumnModuleArticle],
				m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleArticle], TIME_SIZE);
		strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleArticle], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		m_moduleCacheState[line][module][DataCacheColumnModuleArticle] = actual;
	}
}

static KDataCacheModuleColumn kmodule_columnForDataId(KEdcpDataId dataId, bool boardB)
{
	KDataCacheModuleColumn column;

	if (dataId == DCP_DATAID + DCP_DATAID_MOD_LOG_ON)
		return (boardB) ? ExtendedColumnModuleDeviceClass2 : DataCacheColumnModuleDeviceClass;

	if (dataId == DCP_DATAID + DCP_DATAID_MOD_DEVICEID_SOFTWAREID)
		return DataCacheColumnModuleChannelNumber;

	if (dataId == EDCP_DATAID_MOD_FACTORY_FLASH_LOG_ON)
		return DataCacheColumnModuleIdentifierList;

	for (column = 0; column < DataCacheColumnModuleSize; ++column) {
		if (m_moduleItemProperties[column].dataId == dataId)
			return column;
	}

	for (column = DataCacheColumnModuleSize + 1; column < ExtendedColumnModuleSize; ++column) {
		if (m_moduleItemExtendedProperties[column - DataCacheColumnModuleSize - 1].dataId == dataId)
			return column;
	}

	return DataCacheColumnModuleSize;
}

static KCanMessage oldMsg;
void kmodule_setCanMessageValue(const KCanMessage *rxMsg)
{
	bool boardB = 0;
	const uint8_t line = rxMsg->id.bit.addressOffset;
	const uint8_t module = (m_crateCacheValue[0][0].backPlaneType & (1 << line)) ?
				kcan_moduleAddressToSlotNumber(line, rxMsg->id.bit.address, &boardB) : rxMsg->id.bit.address;
	static int counterEvent = 0;

	if (
			m_timeOutBackPlaneType <= TIME_OUT_ECH4XA_BACK_PLANE &&
			m_crateCacheState[0][0][ExtendedColumnCrateBackPlaneType] != actual)
		return;

	memcpy(&oldMsg, rxMsg, sizeof(KCanMessage));
	if (rxMsg->dataId & EDCP_DATAID_CHANNEL_CODE) {
		const uint8_t channel = rxMsg->data[0] + boardB * m_moduleCacheValue[line][module].channelNumber;
		uint32_t value = 0;
		KDataCacheChannelColumn column = kchannel_columnForDataId(rxMsg->dataId & ~EDCP_DATAID_GROUP_CODE);

		if (rxMsg->bit.dlc > 7)
			return;

		if (rxMsg->bit.dlc <= 3) {
			m_channelCache[line][module][channel][column].state = actual;
			m_moduleCacheValue[line][module].timeOut.answers--;

			return;
		}

		memrcpy(&value, &rxMsg->data[1], rxMsg->bit.dlc - 3);
		kchannel_setReceivedValue(value, line, module, channel, column);

	} else {
		KDataCacheModuleColumn column = kmodule_columnForDataId(rxMsg->dataId, boardB);
		uint16_t dataId = (column == DataCacheColumnModuleIdentifierList) ? (DCP_DATAID + DCP_DATAID_MOD_LOG_ON) : rxMsg->dataId;

		if (dataId & EDCP_DATAID_MODULE_CODE) {
			if (column == DataCacheColumnModuleFirmwareName || column == DataCacheColumnModuleHighVoltageOk) {
				char string[EDCP_MAX_STRING_LENGTH];
				uint8_t length = rxMsg->bit.dlc - 2;
				strncpy(string, (char *)&rxMsg->data[0], length);
				string[length] = 0;
				kmodule_setReceivedStringValue(string, line, module, column);
			} else if (column == DataCacheColumnModuleArticle) {
				kmodule_setReceivedStringGroupValue((char *)&rxMsg->data[0], rxMsg->bit.dlc - 3, line, module, column);
			} else {
				uint32_t value = ((uint32_t)rxMsg->data[0] << 8) | ((uint32_t)rxMsg->data[1] << 0);
				if (rxMsg->bit.dlc == 3) {
					value = rxMsg->data[0];
				} else if (rxMsg->bit.dlc == 6) {
					value <<= 16;
					value |= ((uint32_t)rxMsg->data[2] << 8) | ((uint32_t)rxMsg->data[3] << 0);
				}
				kmodule_setReceivedValue(value, line, module, column, NO_BIT_ARRAY_INDEX);
			}
		} else if ((dataId & (EDCP_DATAID_CHANNEL_CODE + EDCP_DATAID_GROUP_CODE)) ==
									 EDCP_DATAID_GROUP_CODE) {
			uint32_t value = ((uint32_t)rxMsg->data[1] << 24) +
											((uint32_t)rxMsg->data[2] << 16) +
												((uint32_t)rxMsg->data[3] << 8) +
													((uint32_t)rxMsg->data[4]);
			kmodule_setReceivedValue(value, line, module, column, rxMsg->data[0]);
		} else if (dataId == DCP_DATAID + DCP_DATAID_MOD_LOG_ON) {
			uint32_t value = rxMsg->data[1];		//DeviceClass
			kmodule_setReceivedValue(value, line, module, column, 0);
		} else if (dataId == DCP_DATAID + DCP_DATAID_MOD_DEVICEID_SOFTWAREID) {
			uint32_t value = rxMsg->data[5];		//ChannelNumber
			kmodule_setReceivedValue(value, line, module, column, 0);
		}
	}

	if (
			(
				m_moduleCacheValue[line][module].timeOut.canId.bit.address == rxMsg->id.bit.address ||
				(
					kcan_isDoubleModule(line, module) &&
					module == kcan_moduleAddressToSlotNumber(line, m_moduleCacheValue[line][module].timeOut.canId.bit.address, &boardB)
				)
			) &&
			m_moduleCacheValue[line][module].timeOut.answers

	) {
		pthread_mutex_lock(&m_mutexAnswers);
		m_moduleCacheValue[line][module].timeOut.answers--;
		pthread_mutex_unlock(&m_mutexAnswers);
	} else {
		writeLogEvent("Event", 1,
					  ++counterEvent, m_moduleCacheValue[line][module].timeOut.answers);

		writeLogEvent("Event", 2,
					  counterEvent, m_moduleCacheValue[line][module].timeOut.cacheState);

		writeLogEvent("Event", 3,
					 counterEvent, state_waitForAnswer);

		writeLogEvent("Event", 4,
					  counterEvent, m_moduleCacheValue[line][module].timeOut.canId.bit.address);

		writeLogEvent("Event", 5,
					   counterEvent, rxMsg->id.bit.address);
	}
}

void kmodule_addAnswers(uint16_t line, uint16_t module, uint16_t answers)
{
	pthread_mutex_lock(&m_mutexAnswers);
	m_moduleCacheValue[line][module].timeOut.answers += answers;
	pthread_mutex_unlock(&m_mutexAnswers);
}

static KCanMessage kmodule_canMessageValueR4(uint16_t line, uint16_t module, bool boardB, uint16_t column, float value)
{
	uint8_t data[4];
	KCanId id = { 0 };
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	id.bit.address = module;
	id.bit.passive = 1;
	memcpy(&data[0], &value, 4);
	KCanMessage msg = kcan_createStandardEdcpMessage(id, boardB, extendedId, m_moduleItemProperties[column].dataId, 6, data);
	//kmodule_addAnswers(line, module, 0);

	return msg;
}

static KCanMessage kmodule_canMessageValueUI2(uint16_t line, uint16_t module, bool boardB, uint16_t column, uint16_t value)
{
	uint8_t data[2];
	KCanId id = { 0 };
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	id.bit.address = module;
	id.bit.passive = 1;
	memcpy(&data[0], &value, 2);
	KCanMessage msg = kcan_createStandardEdcpMessage(id, boardB, extendedId, m_moduleItemProperties[column].dataId, 4, data);

	return msg;
}

static KCanMessage kmodule_canMessageValueString(uint16_t line, bool boardB, uint16_t module, uint16_t column, char *string)
{
	KCanId id = { 0 };
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	id.bit.address = module;
	id.bit.passive = 1;
	KCanMessage msg = kcan_createStandardEdcpMessage(id, boardB, extendedId, m_moduleItemProperties[column].dataId, 8, 0);
	memcpy(msg.data, string, 6);

	return msg;
}

static KCanMessage kmodule_requestMessageValue(uint16_t line, uint16_t module, bool boardB, uint16_t column)
{
	KCanId id = { 0 };
	bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
	id.bit.passive = 1;
	id.bit.address = module;
	KCanMessage msg = kcan_createStandardEdcpMessage(id, boardB, extendedId, m_moduleItemProperties[column].dataId, 2, 0);
	msg.id.bit.readRequest = 1;
	kmodule_addAnswers(line, module, 1);
	m_moduleCacheValue[line][module].timeOut.time = 50;

	return msg;
}

static bool kmodule_isSetCacheValueState(uint16_t line, uint16_t module, ElementState *elementState, CacheState state, bool *boardB)
{
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !boardB) {
		*boardB = true;
		return false;
	}

	*boardB = false;
	*elementState = state;
	return true;
}

static bool kmodule_isDoubleModule(uint16_t line, uint16_t module)
{
	return m_moduleCacheValue[line][module].logOn[0] && m_moduleCacheValue[line][module].logOn[BoardBModule];
}

static void kmodule_counterSetValueIncrement(uint16_t line, uint16_t module)
{
	if (m_moduleCacheValue[line][module].serialNumber < 999999)
		return;

	m_moduleCacheValue[line][module].counterSetValuesChanged++;
}

uint16_t kmodule_channelNumber(uint16_t line, uint16_t module)
{
	return m_moduleCacheValue[line][module].channelNumber * (1 + m_moduleCacheValue[line][module].logOn[BoardBModule]);
}

static bool kmodule_writeData(uint16_t line, uint16_t module, KCanMessage *msg)
{
	static bool boardB = 0;
	uint16_t channel;
	uint16_t channelNumber = kmodule_channelNumber(line, module);

	for (channel = 0; channel < channelNumber; channel++) {
		uint16_t column;
		for (column = DataCacheColumnChannelControl; column <= DataCacheColumnChannelVctCoefficient; column++) {
			ElementState localState = m_channelCache[line][module][channel][column].state;
			if (localState == newValueLowPriority) {
				kchannel_setOffInConsequenceFromCrateControl(line, module, channel, column);
				*msg = kchannel_canMessageValue(line, module, channel, column);
				kmodule_counterSetValueIncrement(line, module);
				m_channelCache[line][module][channel][column].state = ksystem_requestDelay();
				return true;
			}
			if (localState >= waitForRequest1 && localState <= lastWaitForRequest) {
				++m_channelCache[line][module][channel][column].state;
			}
			if (localState == request) {
				*msg = kchannel_requestMessageValue(line, module, channel, column);
				m_channelCache[line][module][channel][column].state = waitForInitialData;
				return true;
			}
		}
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleVoltageRampSpeed] == newValueLowPriority) {
		*msg = kmodule_canMessageValueR4(line, module, boardB, DataCacheColumnModuleVoltageRampSpeed,
										 m_moduleCacheValue[line][module].voltageRampSpeedSet);
		kmodule_counterSetValueIncrement(line, module);

		if (kmodule_isDoubleModule(line, module) && !boardB) {
			boardB = true;
			return true;
		}

		boardB = false;
		kmodule_isSetCacheValueState(line, module,
									 &m_moduleCacheState[line][module][DataCacheColumnModuleVoltageRampSpeed],
									 request, &boardB);
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleVoltageRampSpeed] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleVoltageRampSpeed);
		kmodule_isSetCacheValueState(line, module,
									 &m_moduleCacheState[line][module][DataCacheColumnModuleVoltageRampSpeed],
									 waitForInitialData, &boardB);
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleCurrentRampSpeed] == newValueLowPriority) {
		*msg = kmodule_canMessageValueR4(line, module, boardB, DataCacheColumnModuleCurrentRampSpeed,
										 m_moduleCacheValue[line][module].currentRampSpeedSet);
		kmodule_counterSetValueIncrement(line, module);

		if (kmodule_isDoubleModule(line, module) && !boardB) {
			boardB = true;
			return true;
		}

		boardB = false;
		kmodule_isSetCacheValueState(line, module,
									 &m_moduleCacheState[line][module][DataCacheColumnModuleCurrentRampSpeed],
									 request, &boardB);
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleCurrentRampSpeed] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleCurrentRampSpeed);
		kmodule_isSetCacheValueState(line, module,
									 &m_moduleCacheState[line][module][DataCacheColumnModuleCurrentRampSpeed],
									 waitForInitialData, &boardB);
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleControl] == newValueLowPriority) {
		*msg = kmodule_canMessageValueUI2(line, module, boardB, DataCacheColumnModuleControl,
										  m_moduleCacheValue[line][module].controlSet);
		kmodule_counterSetValueIncrement(line, module);

		if (kmodule_isDoubleModule(line, module) && !boardB) {
			boardB = true;
			return true;
		}

		boardB = false;
		m_moduleCacheState[line][module][DataCacheColumnModuleControl] = request;
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleControl] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleControl);
		m_moduleCacheState[line][module][DataCacheColumnModuleControl] = waitForInitialData;
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleEventStatus] == newValueLowPriority) {
		*msg = kmodule_canMessageValueUI2(line, module, boardB, DataCacheColumnModuleEventStatus,
										  m_moduleCacheValue[line][module].eventStatusSet);
		kmodule_counterSetValueIncrement(line, module);

		if (kmodule_isDoubleModule(line, module) && !boardB) {
			boardB = true;
			return true;
		}

		boardB = false;
		m_moduleCacheState[line][module][DataCacheColumnModuleEventStatus] = request;
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleEventStatus] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleEventStatus);
		m_moduleCacheState[line][module][DataCacheColumnModuleEventStatus] = waitForInitialData;
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleEventMask] == newValueLowPriority) {
		*msg = kmodule_canMessageValueUI2(line, module, boardB, DataCacheColumnModuleEventMask,
										  m_moduleCacheValue[line][module].eventMaskSet);
		m_moduleCacheState[line][module][DataCacheColumnModuleEventMask] = request;
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleEventMask] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleEventMask);
		m_moduleCacheState[line][module][DataCacheColumnModuleEventMask] = waitForInitialData;
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleSampleRate] == newValueLowPriority) {
		*msg = kmodule_canMessageValueUI2(line, module, boardB, DataCacheColumnModuleSampleRate,
										  m_moduleCacheValue[line][module].sampleRateSet);
		kmodule_counterSetValueIncrement(line, module);

		if (kmodule_isDoubleModule(line, module) && !boardB) {
			boardB = true;
			return true;
		}

		boardB = false;
		m_moduleCacheState[line][module][DataCacheColumnModuleSampleRate] = request;
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleSampleRate] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleSampleRate);
		m_moduleCacheState[line][module][DataCacheColumnModuleSampleRate] = waitForInitialData;
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleDigitalFilter] == newValueLowPriority) {
		*msg = kmodule_canMessageValueUI2(line, module, boardB, DataCacheColumnModuleDigitalFilter,
										  m_moduleCacheValue[line][module].digitalFilterSet);
		kmodule_counterSetValueIncrement(line, module);

		if (kmodule_isDoubleModule(line, module) && !boardB) {
			boardB = true;
			return true;
		}

		boardB = false;
		m_moduleCacheState[line][module][DataCacheColumnModuleDigitalFilter] = request;
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleDigitalFilter] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleDigitalFilter);
		m_moduleCacheState[line][module][DataCacheColumnModuleDigitalFilter] = waitForInitialData;
		return true;
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleHighVoltageOk] == newValueLowPriority) {
		*msg = kmodule_canMessageValueString(line, module, boardB, DataCacheColumnModuleHighVoltageOk,
											 m_moduleCacheValue[line][module].highVoltageOkSet);
		m_moduleCacheState[line][module][DataCacheColumnModuleHighVoltageOk] = request;
		return true;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleHighVoltageOk] == request) {
		*msg = kmodule_requestMessageValue(line, module, boardB, DataCacheColumnModuleHighVoltageOk);
		m_moduleCacheState[line][module][DataCacheColumnModuleHighVoltageOk] = waitForInitialData;
		return true;
	}

	return false;
}

static bool kmodule_writeDisconnected(uint16_t line, uint16_t module, KCanMessage *msg)
{
	if (m_moduleCacheState[line][module][DataCacheColumnModuleLogOn] == newValueLowPriority) {
		*msg = kcan_setLogOff(line, module, 0);
		m_moduleCacheState[line][module][DataCacheColumnModuleLogOn] = actual;
		return true;
	}

	return false;
}

static bool kmodule_configureRequestMessage(KCanMessage *msg, uint16_t line, uint16_t module, uint16_t dataId, bool boardB)
{
	msg->bit.extendedId = (msg->id.bit.addressOffset = line) ? 1 : 0;
	msg->id.bit.address = (m_crateCacheValue[0][0].backPlaneType & (1 << line)) ?
								(
									(kmodule_isDoubleModule(line, module)) ?
										kcan_slotNumberToDoubleModuleAddress(module, boardB) :
										kcan_slotNumberToSingleModuleAddress(module)
								) : module;
	msg->id.bit.passive = 1;
	msg->id.bit.readRequest = 1;
	msg->bit.busy = 1;
	msg->bit.edcp = 1;
	msg->bit.dlc = 2;
	msg->dataId = dataId;
	if (kcan_isChannelDataId(dataId)) {
		msg->bit.dlc = 5;
		if (m_moduleCacheValue[line][module].channelNumber <= 16) {
			memrcpy(msg->data, &m_moduleCacheValue[line][module].channelMask, 2);
		}
		kmodule_addAnswers(line, module, m_moduleCacheValue[line][module].channelNumber);
		m_moduleCacheValue[line][module].timeOut.time = 500;
	} else if (isGroupDataId(dataId)) {
		if (dataId == EDCP_DATAID_GRP_TEMPERATURE_SENSOR) {
			kmodule_addAnswers(line, module, 1);
				// z.Z. fehlt die Unterstuetzung fuer mehrere Antworten in der KernelLibrary, d.h
				// sobald die Implementation erfolgt ist, kommt zu kurzzeitigen Arbitrierungen
				// bei wenigen derartigen Ereignissen unkritisch
		} else {
			kmodule_addAnswers(line, module, 8);
		}
		m_moduleCacheValue[line][module].timeOut.time = 500;
	} else {
		kmodule_addAnswers(line, module, 1);
		m_moduleCacheValue[line][module].timeOut.time = 50;
	}

	return false;
}

typedef enum ChannelVariableDataHighPriority {
	REQUEST_CHANNEL_VOLTAGE_MEASURE = 0,
	REQUEST_CHANNEL_CURRENT_MEASURE,
	REQUEST_CHANNEL_STATUS,
	REQUEST_CHANNEL_EVENT_STATUS,
	REQUEST_CHANNEL_NUMBER
} ChannelVariableDataHighPriority;


static KCanMessage kmodule_variableDataHighPriority(uint16_t line, uint16_t *module)
{
	KCanMessage msg = EmptyCanMessage;
	static uint16_t request = REQUEST_CHANNEL_VOLTAGE_MEASURE;
	static bool requestBoardB = false;

	switch (request) {

	case REQUEST_CHANNEL_VOLTAGE_MEASURE:
		if (m_line == DataCacheMaxLines) {
			m_line = line;
		}
		if (m_line == line && *module == atoi(m_lineCacheValue[m_line].moduleList)) {
			m_systemCacheValue.cycleCounter++;
			createTimeStamp(m_systemCacheTimeStampLastChanged[DataCacheColumnSystemCycleCounter]);
			strncpy(m_systemCacheQuality[DataCacheColumnSystemCycleCounter], ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		}

		kmodule_configureRequestMessage(&msg, line,*module, EDCP_DATAID_MBR_VOLTAGE_MEASURE, requestBoardB);
		break;

	case REQUEST_CHANNEL_CURRENT_MEASURE:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MBR_CURRENT_MEASURE, requestBoardB);
		break;

	case REQUEST_CHANNEL_STATUS:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MBR_STATUS, requestBoardB);
		break;

	case REQUEST_CHANNEL_EVENT_STATUS:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MBR_EVENT_STATUS, requestBoardB);
		break;

	}

	m_moduleCacheValue[line][*module].timeOut.canId = msg.id;
	m_moduleCacheValue[line][*module].timeOut.dataId = msg.dataId;
	m_moduleCacheValue[line][*module].timeOut.cacheState = state_waitForAnswer;

	if (!m_moduleCacheValue[line][*module].logOn[BoardBModule] || requestBoardB) {
		requestBoardB = false;
		if (++request >= REQUEST_CHANNEL_NUMBER) {
			request = REQUEST_CHANNEL_VOLTAGE_MEASURE;
			(*module)++;
		}
	} else {
		requestBoardB = true;
	}

	return msg;
}

typedef enum ModuleVariableDataLowPriority {
	REQUEST_MODULE_STATUS = 40,
	REQUEST_MODULE_STATUS_2,
	REQUEST_MODULE_EVENT_STATUS,
	REQUEST_MODULE_VOLTAGE_LIMIT,
	REQUEST_MODULE_CURRENT_LIMIT,
	REQUEST_MODULE_TEMPERATURE,
	REQUEST_MODULE_TEMPERATURE_SENSOR,
	REQUEST_MODULE_SUPPLY,
	REQUEST_MODULE_COUNTER_SET_VALUES_CHANGED,
	REQUEST_CHANNEL_TEMPERATURE_SENSOR,
	REQUEST_MODULE_PRIORITY_NUMBER
} ModuleVariableDataLowPriority;

typedef enum KModuleOption {
	ARTICLE_OPTION_SLA = 1,
	ARTICLE_OPTION_SLP = 2,
	ARTICLE_OPTION_T10 = 4,
	ARTICLE_OPTION_VCT = 8,
	ARTICLE_OPTION_VLN = 16
} KModuleOption;

static KCanMessage kmodule_variableDataLowPriority(uint16_t line, uint16_t *module)
{
	KCanMessage msg = EmptyCanMessage;

	switch (m_moduleCacheValue[line][*module].lowPriorityCounter) {
	case REQUEST_MODULE_STATUS:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_STATUS, 0);
		break;

	case REQUEST_MODULE_STATUS_2:
		if (m_systemCacheValue.logLevel==255 && m_moduleCacheValue[line][*module].serialNumber > 999999) {
			kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_STATUS2, 0);
			break;
		} else {
			m_moduleCacheValue[line][*module].lowPriorityCounter++;
		}

	case REQUEST_MODULE_EVENT_STATUS:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_EVENT_STATUS, 0);
		break;

	case REQUEST_MODULE_VOLTAGE_LIMIT:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_VOLTAGE_LIMIT_REL, 0);
		break;

	case REQUEST_MODULE_CURRENT_LIMIT:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_CURRENT_LIMIT_REL, 0);
		break;

	case REQUEST_MODULE_TEMPERATURE:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_TEMPERATURE, 0);
		break;

	case REQUEST_MODULE_TEMPERATURE_SENSOR:
		if (m_moduleCacheValue[line][*module].serialNumber > 999999) {
			kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_GRP_TEMPERATURE_SENSOR, 0);
			break;
		}

		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_SUPPLY24, 0);
		break;

	case REQUEST_MODULE_SUPPLY:
		if (m_moduleCacheValue[line][*module].serialNumber > 999999) {
			kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_GRP_SUPPLY_MEASURE, 0);
			break;
		}

		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_SUPPLY5, 0);
		break;

	case REQUEST_MODULE_COUNTER_SET_VALUES_CHANGED:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MOD_COUNTER_SET_VALUES_CHANGED, 0);
		if ((m_moduleCacheValue[line][*module].option & ARTICLE_OPTION_VCT) == 0) {
			m_moduleCacheValue[line][*module].lowPriorityCounter++;
		}

		break;

	case REQUEST_CHANNEL_TEMPERATURE_SENSOR:
		kmodule_configureRequestMessage(&msg, line, *module, EDCP_DATAID_MBR_TEMPERATURE_EXTERNAL, 0);
		break;

	default:
		m_moduleCacheValue[line][*module].lowPriorityCounter = 0;
		return msg;
	}

	m_moduleCacheValue[line][*module].timeOut.canId = msg.id;
	m_moduleCacheValue[line][*module].timeOut.dataId = msg.dataId;
	m_moduleCacheValue[line][*module].timeOut.cacheState = state_waitForAnswer;

	return msg;
}

static KCanMessage kmodule_variableData(uint16_t line, uint16_t *module)
{
	if (++m_moduleCacheValue[line][*module].lowPriorityCounter < REQUEST_MODULE_STATUS)
		return kmodule_variableDataHighPriority(line, module);

	return kmodule_variableDataLowPriority(line, module);
}

typedef enum KArticleLength {
	ARTICLE_LENGTH_SYSTEM_MODULE = 19,
	ARTICLE_LENGTH_BIPOLARE_MODULE = 18,
	ARTICLE_LENGTH_MIX_MODULE = 14
} KArticleLength;

static uint16_t kmodule_articleOption(uint16_t line, uint16_t module)
{
	uint16_t length;
	uint8_t pos;
	char articleOption[4] = { 0 };

	length = strlen(m_moduleCacheValue[line][module].article);
	if (length == ARTICLE_LENGTH_SYSTEM_MODULE) {
		pos = 12;
	} else if (length == ARTICLE_LENGTH_BIPOLARE_MODULE) {
		pos = 11;
	} else if (length == ARTICLE_LENGTH_MIX_MODULE) {
		pos = 9;
	} else {
		return 0;
	}
	memcpy(articleOption, &m_moduleCacheValue[line][module].article[pos], 3);

	return atoi(articleOption);
}

static bool kmodule_isCacheInitialized(uint16_t line, uint16_t module, KCanMessage *msg)
{
	static bool requestVoltageSetBoardB = false;
	static bool requestCurrentSetBoardB = false;
	static bool requestVoltageNominalBoardB = false;
	static bool requestCurrentNominalBoardB = false;
	static bool requestControlBoardB = false;

	if (m_crateCacheValue[0][0].backPlaneType & (1 << line)) {
		if (m_moduleCacheValue[line][module].logOn[BoardBModule]) {
			snprintf(m_moduleCacheValue[line][module].identifierList,
					 sizeof(m_moduleCacheValue[line][module].identifierList),
					 "0x%04x,0x%04x",
					 0x200 + (kcan_slotNumberToDoubleModuleAddress(module, 0) << 3),
					 0x200 + (kcan_slotNumberToDoubleModuleAddress(module, 1) << 3));
		} else {
			snprintf(m_moduleCacheValue[line][module].identifierList,
					 sizeof(m_moduleCacheValue[line][module].identifierList),
					 "0x%04x",
					 0x200 + (kcan_slotNumberToSingleModuleAddress(module) << 3));
		}
	} else {
		snprintf(m_moduleCacheValue[line][module].identifierList,
				 sizeof(m_moduleCacheValue[line][module].identifierList),
				 "0x%04x",
				 0x200 + (module << 3));
	}

	if (m_moduleCacheState[line][module][DataCacheColumnModuleStatus] < actual) {
		requestVoltageSetBoardB = requestCurrentSetBoardB = requestVoltageNominalBoardB = requestCurrentNominalBoardB = false;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_STATUS, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleControl] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_CONTROL, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleEventStatus] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_EVENT_STATUS, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleEventMask] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_EVENT_MASK, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleSampleRate] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_ADC_SAMPLES_PER_SECOND, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleDigitalFilter] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_DIGITAL_FILTER, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleVoltageRampSpeed] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_VOLTAGE_RAMP_SPEED, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleCurrentRampSpeed] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_CURRENT_RAMP_SPEED, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleVoltageLimit] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_VOLTAGE_LIMIT_REL, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleCurrentLimit] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_CURRENT_LIMIT_REL, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleTemperature] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_TEMPERATURE, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleSerialNumber] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_SERIAL_NUMBER, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleChannelNumber] < actual) {
//		return kdatacache_configureRequestMessage(msg, module, EDCP_DATAID_MOD_CHANNEL_NUMBER);
		uint8_t data[1] = { 0 };
		KCanId id = { 0 };
		bool extendedId = (id.bit.addressOffset = line) ? 1 : 0;
		id.bit.address = module;
		id.bit.passive = 1;
		*msg = kcan_createStandardDcpMessage(id, false, extendedId, DCP_DATAID_MOD_DEVICEID_SOFTWAREID, 1, data);
		msg->id.bit.readRequest = 1;
		kmodule_addAnswers(line, module, 1);
		m_moduleCacheValue[line][module].timeOut.time = 50;
		return false;
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleFirmwareRelease] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_FIRMWARE_RELEASE, 0);
	}
	if (m_moduleCacheState[line][module][DataCacheColumnModuleFirmwareName] < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_FIRMWARE_NAME, 0);
	}
	if (
			m_moduleCacheState[line][module][DataCacheColumnModuleHighVoltageOk] < actual &&
			(m_moduleCacheValue[line][module].deviceClass == DC_MICC || m_moduleCacheValue[line][module].deviceClass == DC_MICP)
	) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_CONFIGURE_HIGH_VOLTAGE_OK, 0);
	}

	if (
			m_moduleCacheValue[line][module].serialNumber > 999999 &&
			m_moduleCacheState[line][module][DataCacheColumnModuleArticle] < actual
	) {
		m_moduleCacheState[line][module][DataCacheColumnModuleArticle] = actual;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MOD_ARTICLE_DESCRIPTION, 0);
	}

	if (m_channelCache[line][module][0][DataCacheColumnChannelVoltageSet].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VOLTAGE_SET, 0);
	}
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !requestVoltageSetBoardB) {
		requestVoltageSetBoardB = true;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VOLTAGE_SET, requestVoltageSetBoardB);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelCurrentSet].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CURRENT_SET, 0);
	}
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !requestCurrentSetBoardB) {
		requestCurrentSetBoardB = true;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CURRENT_SET, requestCurrentSetBoardB);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelVoltageMeasure].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VOLTAGE_MEASURE, 0);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelCurrentMeasure].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CURRENT_MEASURE, 0);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelVoltageBounds].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VOLTAGE_BOUNDS, 0);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelCurrentBounds].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CURRENT_BOUNDS, 0);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelVoltageNominal].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VOLTAGE_NOMINAL, 0);
	}
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !requestVoltageNominalBoardB) {
		requestVoltageNominalBoardB = true;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VOLTAGE_NOMINAL, requestVoltageNominalBoardB);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelCurrentNominal].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CURRENT_NOMINAL, 0);
	}
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !requestCurrentNominalBoardB) {
		requestCurrentNominalBoardB = true;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CURRENT_NOMINAL, requestCurrentNominalBoardB);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelStatus].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_STATUS, 0);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelControl].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CONTROL, 0);
	}
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !requestControlBoardB) {
		requestControlBoardB = true;
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_CONTROL, requestControlBoardB);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelEventStatus].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_EVENT_STATUS, 0);
	}
	if (m_channelCache[line][module][0][DataCacheColumnChannelEventMask].state < actual) {
		return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_EVENT_MASK, 0);
	}
	if (m_moduleCacheValue[line][module].serialNumber > 999999) {

		m_moduleCacheValue[line][module].option = kmodule_articleOption(line, module);

		if (m_channelCache[line][module][0][DataCacheColumnChannelDelayedTripAction].state < actual) {
			return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_DELAYED_TRIP_ACTION, 0);
		}
		if (m_channelCache[line][module][0][DataCacheColumnChannelDelayedTripTime].state < actual) {
			return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_DELAYED_TRIP_TIME, 0);
		}
		if (m_channelCache[line][module][0][DataCacheColumnChannelExternalInhibitAction].state < actual) {
			return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_EXTERNAL_INHIBIT_ACTION, 0);
		}
		if (m_moduleCacheValue[line][module].option & ARTICLE_OPTION_VCT) {
			if (m_channelCache[line][module][0][DataCacheColumnChannelTemperatureExternal].state < actual) {
				return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_TEMPERATURE_EXTERNAL, 0);
			}
			if (m_channelCache[line][module][0][DataCacheColumnChannelVctCoefficient].state < actual) {
				return kmodule_configureRequestMessage(msg, line, module, EDCP_DATAID_MBR_VCT_PARAMETER, 0);
			}
		}
	}

	m_moduleCacheValue[line][module].timeOut.answers = 0;
	return true;
}

static bool kmodule_setElementStateForAll(uint16_t line, uint16_t module, ElementState state)
{
	uint16_t channel;
	KDataCacheModuleColumn column;

	if (state == waitForInitialData) {
		m_moduleCacheState[line][module][DataCacheColumnModuleAlive] = waitForInitialData;
	}

	for (column = DataCacheColumnModuleStatus; column < DataCacheColumnModuleSize; column++) {
		m_moduleCacheState[line][module][column] = state;
		if (state == communicationBad) {
			strncpy(m_moduleCacheQuality[line][module][column], ISEG_ITEM_QUALITY_COMMUNICATION_BAD, QUALITY_SIZE);
		}
	}

	uint16_t channelNumber = kmodule_channelNumber(line, module);
	for (channel = 0; channel < channelNumber; channel++) {
		KDataCacheChannelColumn column;

		for (column = DataCacheColumnChannelStatus; column < DataCacheColumnChannelSize; column++) {
			m_channelCache[line][module][channel][column].state = state;
			if (state == communicationBad) {
				strncpy(m_channelCache[line][module][channel][column].quality, ISEG_ITEM_QUALITY_COMMUNICATION_BAD, QUALITY_SIZE);
			}
		}
	}

	return true;
}

static bool kmodule_isTimedOut(uint16_t line, uint16_t module)
{
	if (m_moduleCacheValue[line][module].timeOut.time) {
		m_moduleCacheValue[line][module].timeOut.time--;
	}

	return !m_moduleCacheValue[line][module].timeOut.time && m_moduleCacheValue[line][module].timeOut.answers;
}

void kmodule_setAlive(bool alive, uint16_t line, uint16_t module)
{
	if (alive != m_moduleCacheValue[line][module].alive) {
		char *moduleList = m_lineCacheValue[line].moduleList;
		int byteWritten = 0;

		createTimeStamp(m_lineCacheTimeStampLastChanged[line][DataCacheColumnLineModuleNumber]);
		strncpy(m_lineCacheQuality[line][DataCacheColumnLineModuleNumber],
					ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);

		createTimeStamp(m_systemCacheTimeStampLastChanged[DataCacheColumnSystemModuleNumber]);
		strncpy(m_systemCacheQuality[DataCacheColumnSystemModuleNumber], ISEG_ITEM_QUALITY_OK,
					QUALITY_SIZE);

		m_moduleCacheValue[line][module].alive = alive;
		strncpy(m_moduleCacheQuality[line][module][DataCacheColumnModuleAlive],
				ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
		createTimeStamp(m_moduleCacheTimeStampLastRefreshed[line][module][DataCacheColumnModuleAlive]);

		*moduleList = 0;
		for (module = 0; module < DataCacheMaxModules; module++) {
			if (m_moduleCacheValue[line][module].alive) {
				byteWritten += snprintf(&moduleList[byteWritten],
										VALUE_SIZE - byteWritten,
										"%lu,", (unsigned long)module);
			}
		}
		if (byteWritten)
			moduleList[byteWritten - 1] = 0;
		createTimeStamp(m_lineCacheTimeStampLastRefreshed[line][DataCacheColumnLineModuleList]);
		createTimeStamp(m_lineCacheTimeStampLastChanged[line][DataCacheColumnLineModuleList]);
		strncpy(m_lineCacheQuality[line][DataCacheColumnLineModuleList],
				ISEG_ITEM_QUALITY_OK, QUALITY_SIZE);
	}
}

bool kmodule_isSetCacheTimeOutState(uint16_t line, uint16_t module, CacheState state, bool *boardB)
{
	if (m_moduleCacheValue[line][module].logOn[BoardBModule] && !(*boardB)) {
		*boardB = true;
		return false;
	}

	*boardB = false;
	m_moduleCacheValue[line][module].timeOut.cacheState = state;
	return true;
}

KCanMessage kmodule_handler(uint32_t *preConfiguredCounter, KCollectData *collectData)
{
	KCanMessage msg = EmptyCanMessage;
	static uint16_t line = 0;
	static uint16_t module = 0;
	static uint8_t notConnectedModules = 0;
	static bool boardB = false;

	switch (m_moduleCacheValue[line][module].timeOut.cacheState) {
	case state_notLogged:
		if (
				kcan_logged(line, module) &&
				m_timeOutBackPlaneType > TIME_OUT_MODULE_LOG_ON &&
				(
					(
						m_crateCacheValue[0][0].alive &&
						m_crateCacheState[0][0][ExtendedColumnCrateBackPlaneType] == actual
					) ||
					(
						!m_crateCacheValue[0][0].alive
					)
				)
		) {
			msg = kcan_setLogOn(line, module, boardB);
			kmodule_setElementStateForAll(line, module, waitForInitialData);
			kmodule_isSetCacheTimeOutState(line, module, state_configure, &boardB);
			memset(m_moduleCacheValue[line][module].article, 0, EDCP_GROUP_MAX_STRING_LENGTH);
			break;
		}

		if (kmodule_writeDisconnected(line, module, &msg))
			return msg;

		do {
			if (module == DataCacheMaxModules) {
				if (notConnectedModules == DataCacheMaxModules) {
					++ *preConfiguredCounter;
				}

				notConnectedModules = 0;
				break;
			}

			if (m_moduleCacheValue[line][module].timeOut.cacheState == state_logOff)
				break;

			notConnectedModules++;
		} while (!kcan_logged(line, ++module));


		break;

	case state_logOff:
		msg = kcan_setLogOff(line, module, boardB);
		if (boardB) {
			m_moduleCacheValue[line][module].logOn[BoardBModule] = false;
			m_moduleCacheValue[line][module].logOn[BoardASingleModule] = false;
			m_moduleCacheValue[line][module].logOn[BoardBSingleModule1] = false;
			m_moduleCacheValue[line][module].logOn[BoardBSingleModule2] = false;
		} else if (!m_moduleCacheValue[line][module].logOn[BoardBModule]) {
			m_moduleCacheValue[line][module].logOn[BoardASingleModule] = false;
			m_moduleCacheValue[line][module].logOn[BoardBSingleModule1] = false;
			m_moduleCacheValue[line][module].logOn[BoardBSingleModule2] = false;
		}
		m_moduleCacheValue[line][module].downCounterLogOn = 5;
		pthread_mutex_lock(&m_mutexAnswers);
		m_moduleCacheValue[line][module].timeOut.answers = 0;
		pthread_mutex_unlock(&m_mutexAnswers);
		kmodule_isSetCacheTimeOutState(line, module, state_notLogged, &boardB);
		break;

	case state_configure:
		msg = kcan_createAntiCollisionMessage(line, module, boardB);
		kmodule_isSetCacheTimeOutState(line, module, state_uninitialized, &boardB);
		break;

	case state_uninitialized:
		if (!kmodule_isCacheInitialized(line, module, &msg)) {
			m_moduleCacheValue[line][module].timeOut.canId = msg.id;
			m_moduleCacheValue[line][module].timeOut.dataId = msg.dataId;
			m_moduleCacheValue[line][module].timeOut.cacheState = state_waitForAnswer;
			break;
		}

		kmodule_setAlive(true, line, module);
		m_moduleCacheValue[line][module].updated = true;
		m_moduleCacheValue[line][module].timeOut.cacheState = state_requestVariableData;
		break;

	case state_waitForAnswer:
		if (kmodule_isTimedOut(line, module)) {
			kloggerfifo_setMessage(LogErrors, "Stop module at ", line, module);
			kmodule_setAlive(false, line, module);
			if (
					m_line == line &&
					m_lineCacheValue[line].moduleList[0] == 0
			) {
				m_line = DataCacheMaxLines;
			}

			if (kmodule_isSetCacheTimeOutState(line, module, state_logOff, &boardB)) {
				kmodule_setElementStateForAll(line, module, communicationBad);
			}
			break;
		}

		if (kcan_cacheAllAnwersReceived(line, module)) {
			if (!kcan_cacheAlive(line, module) || !kcan_cacheSetValuesUpdated(line, module)) {
				m_moduleCacheValue[line][module].timeOut.cacheState = state_uninitialized;
				break;
			}

			m_moduleCacheValue[line][module].timeOut.cacheState = state_requestVariableData;
		}
		break;

	case state_requestVariableData:
		if (kmodule_writeData(line, module, &msg)) {
			m_moduleCacheValue[line][module].timeOut.cacheState = state_requestVariableData;
			break;
		}

		msg = kmodule_variableData(line, &module);
		break;
	}

	if (module < DataCacheMaxLines && m_moduleCacheValue[line][module].timeOut.cacheState != state_notLogged)
		*preConfiguredCounter = 0;

	if (module == DataCacheMaxModules) {
		module = 0;
		notConnectedModules = 0;
		if (++line == DataCacheMaxLines) {
			line = 0;
			*collectData = CollectCrateData;
		}
	}

	return msg;
}
