#include "isegapi.h"

#include <pthread.h>
#include <unistd.h>

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>

void RunAllTests(void);

static void getLine(const char *msg, char *buf, int bufSize)
{
	printf("%s: ", msg);

	fgets(buf, bufSize, stdin);
	char *lf = strchr(buf, '\n');
	if (*lf)
		*lf = 0;
}

static void getItem(const char *name)
{
	char buf[255];
	char *p_pos;
	getLine("Enter item to get", buf, sizeof(buf));

	if (strchr(buf, '*')) {
		uint16_t pos;
		uint8_t separatorNumber = 0;
		uint8_t wild_card = 0;
		for (pos = 0; pos < strlen(buf); pos++) {
			if (buf[pos] == '.')
				separatorNumber++;
			if (buf[pos] == '*') {
				switch (separatorNumber) {
				case 1:
					wild_card++;
					break;

				case 2:
					if (wild_card)
						wild_card++;

					wild_card++;
					break;
				}
			}
		}

		if (separatorNumber == 3) {
			char buf1[255], buf2[255];

			char moduleList[VALUE_SIZE];
			char *p_mod;
			if (wild_card == 3) {
				strncpy(buf1, "0.ModuleList", FULLY_QUALIFIED_OBJECT_SIZE);
				IsegItem item = iseg_getItem(name, buf1);
				strncpy(moduleList, item.value, VALUE_SIZE);
				p_mod = moduleList;
			}

			do {
				strncpy(buf1,buf, FULLY_QUALIFIED_OBJECT_SIZE);

				if (wild_card == 3) {
					p_pos = strchr(buf1, '*');
					*p_pos = *p_mod;
					p_mod += 2;
				}

				strncpy(buf2, buf1, FULLY_QUALIFIED_OBJECT_SIZE);
				p_pos = strrchr(buf2, '*');
				strncpy(p_pos, "ChannelNumber", FULLY_QUALIFIED_OBJECT_SIZE);
				IsegItem item = iseg_getItem(name, buf2);

				uint16_t chhannelNumber = atoi(item.value);
				uint16_t channel;
				for (channel = 0; channel < chhannelNumber; channel++) {
					IsegItem item;
					strncpy(buf2,buf1, FULLY_QUALIFIED_OBJECT_SIZE);
					if (channel < 10)
						*strchr(buf2, '*') = channel + '0';
					else {
						p_pos = strchr(buf2, '*');
						*p_pos = (channel/10 + '0');
						*(p_pos + 1) = channel%10 + '0';
						strncpy(p_pos + 2, strrchr(buf1, '*') + 1, FULLY_QUALIFIED_OBJECT_SIZE);
					}
					item = iseg_getItem(name, buf2);
					printf("'%s': '%s' (Q: %s TR: %s TC: %s)\n", item.object, item.value, item.quality,
						   item.timeStampLastRefreshed, item.timeStampLastChanged);
				}
			} while (wild_card == 3 && *p_mod != 0);
		} else if (separatorNumber == 2 && (p_pos = strchr(buf, '['))) {
			uint16_t index;
			for (index = 0; index < VCC_ARRAY_SIZE; index++) {
				*(p_pos + 1) = (index + '0');
				IsegItem item = iseg_getItem(name, buf);
				printf("'%s': '%s' (Q: %s TR: %s TC: %s)\n", item.object, item.value, item.quality,
					   item.timeStampLastRefreshed, item.timeStampLastChanged);
			}
		}
	} else {
		IsegItem item = iseg_getItem(name, buf);
		printf("'%s': '%s' (Q: %s TR: %s TC: %s)\n", item.object, item.value, item.quality,
			   item.timeStampLastRefreshed, item.timeStampLastChanged);
	}
}

static void setItem(const char *name)
{
	IsegItem item = EmptyIsegItem;
	getLine("Enter item to set", item.object, sizeof(item.object));
	getLine("Enter new value", item.value, sizeof(item.value));

	iseg_setItem(name, item.object, item.value);
}

static void getItemProperty(const char *name)
{
	char buf[255];
	getLine("Enter item to get property", buf, sizeof(buf));

	IsegItemProperty property = iseg_getItemProperty(name, buf);
	printf("Object: '%s' (Q: %s)\n", property.object, property.quality);
	printf("DataType: '%s'\n",property.type);
	printf("Unit: '%s'\n", property.unit);
	printf("Access rights: '%s'\n", property.access);
}

static void resetEventStatus(const char *name)
{
	IsegItem item = EmptyIsegItem;
	strncpy(item.object, "0.0.0.EventStatus", FULLY_QUALIFIED_OBJECT_SIZE);
	strncpy(item.value, "128", VALUE_SIZE);
	iseg_setItem(name, item.object, item.value);

	strncpy(item.object, "0.0.0.EventStatus", FULLY_QUALIFIED_OBJECT_SIZE);
	strncpy(item.value, "16", VALUE_SIZE);
	iseg_setItem(name, item.object, item.value);
}

static void setVoltage(const char *name)
{
	uint8_t slot, channel;
	static float value = 100.f;
	static uint8_t setOn = 0;

	setOn ^= 1;
	for (slot = 0; slot < 10; slot++) {
		IsegItem item = EmptyIsegItem;
		sprintf(item.object, "0.%d.VoltageRampSpeed", slot);
		sprintf(item.value, "%f", 20.f);
		if (iseg_setItem(name, item.object, item.value) != ISEG_OK) {
			printf("Fehler: %s", item.object);
		}
		for (channel = 0; channel < 16; channel++) {
			IsegItem item = EmptyIsegItem;
			sprintf(item.object, "0.%d.%d.VoltageSet", slot, channel);
			sprintf(item.value, "%f", value);
			if (iseg_setItem(name, item.object, item.value) != ISEG_OK) {
				printf("Fehler: %s", item.object);
			}
			value += 1.f;

			sprintf(item.object, "0.%d.%d.Control:3", slot, channel);
			sprintf(item.value, "%d", setOn);
			if (iseg_setItem(name, item.object, item.value) != ISEG_OK) {
				printf("Fehler: %s", item.object);
			}
		}
		//usleep(200000);
	}
}

static void getRelease(void)
{
	printf("Version string: %s\n", iseg_getVersionString());
	printf("Version number: %08x\n", iseg_getVersion());

}

int UnityMain(int argc, char* argv[], void (*runAllTests)(void));

int main(int argc, char *argv[])
{
	int  statusDevice;
	char name[12];

	if (argc == 2 && strcmp(argv[1], "--test") == 0)
		return UnityMain(argc, argv, RunAllTests);

	printf("isegHAL Example started\n");

	char interface1[] = "can0";
	if (argc == 2) {
		strcpy(interface1, argv[1]);
	}

	char name1[] = "interface1";
	if ( (statusDevice = iseg_connect(name1, interface1, NULL)) != ISEG_OK) {
		printf("Can not connect to line %s.\n", interface1);
		return EXIT_FAILURE;
	} else {
		strncpy(name, name1, sizeof(name));
	}

//	char interface2[] = "can1";
//	char name2[] = "interface2";
//	if ( (statusDevice = iseg_connect(name2, interface2, NULL)) != ISEG_OK) {
//		printf("Can not connect to line %s.\n", interface2);
//		return EXIT_FAILURE;
//	} else {
//		strncpy(name, name2, sizeof(name));
//	}

	int quit = 0;
	do {
		printf("Press 'q' [ENTER] to quit, 'g' for get and 's' for set command,\n");
		printf("press 'p' for get item properties.");
		char buf[255];
		fgets(buf, sizeof(buf), stdin);
		switch (toupper(buf[0])) {
		case 'Q':
			quit = 1;
			break;
		case 'G':
			getItem(name);
			break;
		case 'S':
			setItem(name);
			break;
		case 'P':
			getItemProperty(name);
			break;
		case 'V':
			setVoltage(name);
			break;
		case 'E':
			resetEventStatus(name);
			break;
		case 'R':
			getRelease();
			break;
		}
	} while (!quit);

	iseg_disconnect(name1);
//	iseg_disconnect(name2);

	printf("main stopped (id=%lu)\n", (unsigned long)pthread_self());
	return EXIT_SUCCESS;
}
