#define _MULTI_THREADED

#include "isegcommon.h"
#include "mainthread.h"
#include "receivethread.h"
#include "sendthread.h"
#include "kdatacache.h"

#include <kcan.h>

#include <stdio.h>
#include <stdlib.h>

#include <string.h>

#include <pthread.h>
#include <sys/time.h>
#include <unistd.h>

enum { Debug = 1 };

static int             m_terminate = 0;
static pthread_mutex_t m_mutex = PTHREAD_MUTEX_INITIALIZER;

static pthread_t m_receiveThread;
static pthread_t m_sendThread;
static bool m_startedThreads = false;

static FILE * m_file = NULL;

static void *receiveFunc(void *name)
{
	(void)name;

	while (!m_terminate) {
		KCanMessage rxMsg = receivethread_waitForMessage();

		if (rxMsg.bit.busy) {
			pthread_mutex_lock(&m_mutex);
			kdatacache_setCanMessageValue(&rxMsg);
			pthread_mutex_unlock(&m_mutex);
		}
	}

	pthread_exit((void *)pthread_self());
}

static void *sendFunc(void *name)
{
	static uint16_t loop = 0;
	(void)name;

	while (!m_terminate || !kdatacache_isDisconnected()) {
		struct timeval timeout = { 0, 1000 };
		select(0, 0, 0, 0, &timeout);

		pthread_mutex_lock(&m_mutex);
		KCanMessage msg = kdatacache_handler();
		pthread_mutex_unlock(&m_mutex);

		if (loop < 2000 && ++loop == 2000) {
			char fname[255];
			int number;
			for (number = 0; number < 1000; number++) {
				snprintf(fname, sizeof(fname), "%s/isegHAL%03d.log",
						 kdatacache_getSystemValueString(DataCacheColumnSystemLogPath, 0), number);
				if ((m_file = fopen(fname, "r")) == NULL) {
					break;
				}

				fclose(m_file);
			}
			m_file = fopen(fname, "wt");
		}

		if (msg.bit.busy)
			sendthread_send(&msg);

		if (m_terminate) {
			kdatacache_disconnect();
		}
	}

	printf("sendFunc(): exit %ld\n", m_receiveThread);
	pthread_exit((void *)pthread_self());
}

int ksocketcan_start(const char *line)
{
	printf("start(): started, id=%ld\n", pthread_self());

	m_terminate = 0;

	if (!sendthread_init(line))
		return 0;

	if (!receivethread_init(line))
		return 0;

	if (m_startedThreads)
		return 1;

	if (pthread_create(&m_sendThread, NULL, &sendFunc, NULL) != 0) {
		fprintf(stderr, "start(): could not create send thread!");
		exit(EXIT_FAILURE);
	}
	printf("start(): Send thread %ld started\n", m_sendThread);

	if (pthread_create(&m_receiveThread, NULL, &receiveFunc, NULL) != 0) {
		fprintf(stderr, "start(): could not create receive thread!");
		exit(EXIT_FAILURE);
	}
	printf("start(): Receive thread %ld started\n", m_receiveThread);

	printf("start(): exit (id=%ld)\n", pthread_self());

	m_startedThreads = true;

	return 1;
}

int ksocketcan_stop(void)
{
	m_terminate = 1;

	pthread_join(m_receiveThread, NULL);
	pthread_join(m_sendThread, NULL);

	m_startedThreads = false;

	if (m_file) {
		fclose(m_file);
		m_file = NULL;
	}

	return EXIT_SUCCESS;
}

void kmutex_lock(void)
{
	pthread_mutex_lock(&m_mutex);
}

void kmutex_unlock(void)
{
	pthread_mutex_unlock(&m_mutex);
}

#define DATE_HOUR_MINUTE_SECOND_SIZE 19
static char *createTimeStampDebug(char *timeStamp)
{
	struct timeval time;
	gettimeofday(&time, NULL);	// UTC

	struct tm *utc = gmtime(&time.tv_sec);
	strftime(timeStamp, DATE_HOUR_MINUTE_SECOND_SIZE + 1, "%Y-%m-%d %H:%M:%S", utc); // yyyy-mm-dd hh:mm:ss
	snprintf(&timeStamp[DATE_HOUR_MINUTE_SECOND_SIZE], TIME_SIZE - DATE_HOUR_MINUTE_SECOND_SIZE, ".%04ld", (long)time.tv_usec/100);

	return timeStamp;
}

void writeLogMessage(const char *messageType, const struct can_frame *frame, int count, uint32_t answers)
{
	int i;
	char message[25];

	if (!m_file)
		return;

	fprintf(m_file, "%s ", createTimeStampDebug(message));
	fprintf(m_file, "%s ", messageType);

	if (frame) {
		fprintf(m_file, "%04X [", frame->can_id);
		for (i = 0; i < frame->can_dlc; ++i)
			fprintf(m_file, " %02X", frame->data[i]);
		fprintf(m_file, "] ");
		fprintf(m_file, "number = %lu expected answers = %lu\n", (unsigned long)count, (unsigned long)answers);
	} else {
		fprintf(m_file, "\n");
	}
	fflush(m_file);
}

void writeLogEvent(const char *messageType, uint32_t level, int count, uint32_t answers)
{
	char message[25];

	if (!m_file)
		return;

	fprintf(m_file, "%02lX ", (unsigned long)level);
	fprintf(m_file, "%s ", createTimeStampDebug(message));
	fprintf(m_file, "%s: ", messageType);
	fprintf(m_file, " ] timout number %lu expected answers %lu\n", (unsigned long)count, (unsigned long)answers);
	fflush(m_file);
}


